# %%
'''Function List:
conv_to_array(value): Convert object into an array.
ts_to_spacepystr(ts):  Convert from utc timestamp to datetime string.
ts_to_spacepydt(ts): Convert from utc timestamp to a datetime object.
create_spacepy(inTime, c1, c2, c3, inCoord, inType): Create a SpacePy Coords
    object and return.
create_astropy(inTime, c1, c2, c3, inCoord, inType): Create as AstroPy SkyCoord
    object and return.
spacepy_spacepy(spacepy_coord, outCoord, outType): Converts given SpacePy
    object to the requested SpacePy coordinate system.
astropy_astropy(astropy_coord, outCoord, outType): Converts given Astropy
    object to the requested AstroPy coordinate system.
spacepy_astropy(spacepy_coord, outCoord, outType): Converts given SpacePy
    object to an AstroPy object in the requested coordinate system.
astropy_spacepy(astropy_coord, outCoord, outType): Converts given Astropy
    object to a SpacePy object in the requested coordinate system.
extract_spacepy(inTime, c1, c2, c3, inCoord, inType, outCoord, outType, newC,
    verbose=False): Extract x, y, z, and units from a SpacePy Coords object and
    return.
extract_astropy(astropy_coord, outType): Extract x, y, z, and units from an
    AstroPy SkyCoord object and return.
ConvertCoord(inTime,c1,c2,c3,inCoord,inType,outCoord,outType,verbose=False):
    Main function to perform coordinate conversions.
ComputeLshell(inTime,c1,c2,c3,inCoordName,inCoordType): Computes L Shell
ConvertCoord_f(inTime,c1,c2,c3,iC1,iT1,iC2,iT2): Wrapper function for
    ConvertCoord to be compatible with integer inputs for coordinate systems
    and types.
'''

import time
import numpy as np
from datetime import datetime as dt
from datetime import timezone
import spacepy.coordinates as spc
from spacepy.coordinates import Coords
from spacepy.time import Ticktock
import astropy.units as u
from astropy.coordinates import SkyCoord


# Set option in spacepy to use ctrans instead of irbempy
spc.DEFAULTS.set_values(use_irbem=False, itol=5)

# Define coordinate lists for AstroPy and SpacePy
astropy_coordlist = ['teme', 'icrs', 'fk5', 'fk4', 'itrs', 'galactic', 'cirs',
                     'tete', 'galactocentric', 'precessedgeocentric',
                     'geocentricmeanecliptic', 'geocentrictrueecliptic',
                     'hcrs', 'barycentricmeanecliptic',
                     'heliocentricmeanecliptic', 'barycentrictrueecliptic',
                     'heliocentrictrueecliptic', 'heliocentriceclipticiau76',
                     'custombarycentricecliptic', 'lsr', 'lsrk', 'lsrd',
                     'galacticlsr', 'fk4noeterms', 'supergalactic']
spacepy_coordlist = ['GDZ', 'GEO', 'GSM', 'GSE', 'SM', 'GEI', 'MAG', 'SPH',
                     'RLL']


def convert_to_array(value):
    '''Convert given object to array.'''

    if isinstance(value, (np.ndarray)):
        return value
    else:
        if isinstance(value, (list)):
            return np.array(value)
        else:
            return np.array([value])


@np.vectorize
def ts_to_spacepystr(ts):
    '''Convert from utc timestamp to datetime string.'''
    return dt.strftime(dt.utcfromtimestamp(int(ts)), '%Y-%m-%dT%H:%M:%S')


# Steve Morley recommended converting to something other than a string to
# improve speed.
@np.vectorize
def ts_to_spacepydt(ts):
    '''Convert from utc timestamp to a datetime object.'''
    return dt.fromtimestamp(ts, tz=timezone.utc)

    '''Testing description and results:
    Given nine spherical coordinate positions...
        sat_time = np.repeat(1068865200, 9)  #UTC timestamps
        sat_lon = np.array([-180.,-180.,-180.,0.,0.,0.,180.,180.,180.])  #deg
        sat_lat = np.array([-90.,0.,90.,-90.,0.,90.,-90.,0.,90.])  #in degrees
        sat_radius = np.repeat(1.063871355909377,9)  #in R_earth
    and converting between every possible coordinate pair,
        coord_list = ['GDZ','GEO','GSM','GSE','SM','GEI','MAG','SPH','RLL']
    the ratio of completion times were compared between the original string
    method, the datetime object method and the direct timestamp method, each
    individually to the original method and also to each other (see table
    below). The ratios were chosen so the slower method was on top with the
    faster method on the bottom to indicate how many times faster the 'new'
    method would be as compared to the 'old' one.

    Avg   Std.Dev. Min   Max
    ---------------------------------
    2.05  0.67     1.00  6.2   ratio = str/dt completion times
    1.46  0.52     0.43  4.4   ratio = str/ts completion times
    1.52  0.69     0.73  4.8   ratio = ts/dt  completion times

    Result: The datetime method is the fastest of the three tested, although
    within uncertainties of the direct timestamp method. In addition, the
    datetime method is never slower than the string method unlike the timestamp
    method. Also, the datetime method was 0.55 (+/- 0.32) ms faster on average
    than the string method, and 0.24 (+/- 0.34) ms faster on average than the
    direct timestamp method. So, the datetime object method is preferred.
    '''


def create_spacepy(inTime, c1, c2, c3, inCoord, inType):
    '''Create a SpacePy Coords object and return.'''

    # pack position arrays in desired format
    if inType == "sph":
        # Need to reorder Kamodo arrays to be consistent with spacepy
        sat_track = [[c3[i], c2[i], c1[i]] for i in range(len(c1))]
    else:  # cartesian
        sat_track = [[c1[i], c2[i], c3[i]] for i in range(len(c1))]

    # create SpacePy coordinate object
    cvals = Coords(sat_track, inCoord, inType)
    tt = Ticktock(ts_to_spacepydt(inTime), 'UTC')  # datetime object method
    cvals.ticks = tt
    return cvals


def create_astropy(inTime, c1, c2, c3, inCoord, inType):
    '''Create as AstroPy SkyCoord object and return.'''

    t = ts_to_spacepydt(inTime)
    if inType == 'car':
        astropy_coord = SkyCoord(c1*u.Rearth, c2*u.Rearth, c3*u.Rearth,
                                 obstime=t, frame=inCoord,
                                 representation_type='cartesian')
    elif inType == 'sph':
        astropy_coord = SkyCoord(c1*u.deg, c2*u.deg, c3*u.Rearth, obstime=t,
                                 frame=inCoord,
                                 representation_type='spherical')
    return astropy_coord


def spacepy_spacepy(spacepy_coord, outCoord, outType):
    '''Converts given SpacePy object to the final SpacePy coordinate system.'''

    new_spacepy_coord = spacepy_coord.convert(outCoord, outType)
    assert spacepy_coord.data.shape == new_spacepy_coord.data.shape
    return new_spacepy_coord


def astropy_astropy(astropy_coord, outCoord, outType):
    '''Converts given Astropy object to the final AstroPy coordinate system.'''

    new_astropy_coord = astropy_coord.transform_to(outCoord)
    if outType == 'car':
        new_astropy_coord.representation_type = 'cartesian'
    elif outType == 'sph':
        new_astropy_coord.representation_type = 'spherical'
    return new_astropy_coord


def spacepy_astropy(spacepy_coord, outCoord, outType):
    '''Converts given SpacePy object to an AstroPy object in the final
    coordinate system.'''

    itrs = Coords.to_skycoord(spacepy_coord)  # always to ITRS cartesian
    astropy_coord = astropy_astropy(itrs, outCoord, outType)
    return astropy_coord


def astropy_spacepy(astropy_coord, outCoord, outType):
    '''Converts given Astropy object to a SpacePy object in the final
    coordinate system.'''

    cvals = Coords.from_skycoord(astropy_coord)  # always to GEO car
    spacepy_coord = spacepy_spacepy(cvals, outCoord, outType)
    return spacepy_coord


def extract_spacepy(inTime, c1, c2, c3, inCoord, inType, outCoord, outType,
                    newC, verbose=False):
    '''Extract x, y, z, and units from a SpacePy Coords object and return.'''

    if outType == "sph":
        # Need to reorder Kamodo arrays to be consistent with spacepy
        zz, yy, xx = newC.data.T
    else:
        xx, yy, zz = newC.data.T

    # check for and correct negative radii or altitudes
    # observed in some conversions to GDZ sph
    # also correct for negative longitudes in SPH sph coordinates
    # (should be 0 to 360 longitude)
    if outType == 'sph':
        idx = np.where(zz < 0.)[0]
        # neg radii error only observed to occur when lat=-90 deg in GDZ
        # also occurs at other times. ignoring those...
        close_to_pole = all(np.round(yy[idx], 1) == -90.) or \
            all(np.round(yy[idx], 1) == 90.)
        if len(idx) > 0 and not close_to_pole:
            print('Negative radii/altitudes detected away from poles.')
        if close_to_pole:
            while len(idx) > 0:
                if verbose:
                    print(f'Shifting {len(idx)} latitudes to avoid negative ' +
                          'radii or altitudes.')
                c2[idx] += 0.000000001  # slightly offset lat from -90 deg
                idx2 = np.where(c2 > 90.)[0]  # check for lat>90 after offset
                if len(idx2) > 0:
                    c2[idx2] -= 0.000000002  # correct in other direction
                sat_track_idx = [[c3[idx[i]], c2[idx[i]], c1[idx[i]]] for i in
                                 range(len(idx))]  # make new sat_track
                cvals_idx = Coords(sat_track_idx, inCoord, inType)
                tt_idx = Ticktock(ts_to_spacepydt(inTime[idx]), 'UTC')
                cvals_idx.ticks = tt_idx
                newC_idx = cvals_idx.convert(outCoord, outType)
                zz[idx], yy[idx], xx[idx] = newC_idx.data.T
                idx = np.where(zz < 0.)[0]
        if outCoord == 'SPH':  # SPH sph lon range should be 0 to 360
            idx = np.where(xx < 0)[0]  # select negative longitudes
            xx[idx] += 360  # fix
    outUnits = newC.units
    if outType == "sph":
        # Need to reorder Kamodo arrays to be consistent with spacepy
        outUnits.reverse()

    # fix spacepy bug in returned units
    if outType == 'sph':
        # check altitude for spherical return units
        if max(zz) > 300.:
            if outUnits[2] != 'km':
                outUnits[2] = 'km'
        elif min(zz) < 100.:
            if outUnits[2] != 'Re':
                outUnits[2] = 'Re'
        # else leave as it is
    else:
        # check cartesian position return units
        rr = np.sqrt(xx**2 + yy**2 + zz**2)
        if max(rr) > 300.:
            if outUnits[0] != 'km':
                outUnits = ['km', 'km', 'km']
        elif min(rr) < 100.:
            if outUnits[0] != 'Re':
                outUnits = ['Re', 'Re', 'Re']
        # else leave as it is

    # change unit Re to R_E for recognition in Kamodo
    for i in range(len(outUnits)):
        if outUnits[i] == 'Re':
            outUnits[i] = 'R_E'

    return xx, yy, zz, outUnits


def extract_astropy(astropy_coord, outType):
    '''Extract x, y, z, units from an AstroPy SkyCoord object and return.'''

    # retrieve converted values by the attribute names, converting to the
    #   standard unit based on car vs sph
    coord_names = list(astropy_coord.representation_component_names.keys())
    if outType == 'car':
        units = ['R_E', 'R_E', 'R_E']
        x = getattr(astropy_coord, coord_names[0]).to(u.Rearth).value
        y = getattr(astropy_coord, coord_names[1]).to(u.Rearth).value
    elif outType == 'sph':
        units = ['deg', 'deg', 'R_E']
        x = getattr(astropy_coord, coord_names[0]).to(u.deg).value
        y = getattr(astropy_coord, coord_names[1]).to(u.deg).value
    z = getattr(astropy_coord, coord_names[2]).to(u.Rearth).value
    return x, y, z, units


def ConvertCoord(inTime, c1, c2, c3, inCoord, inType, outCoord, outType,
                 verbose=False):
    """
    This function uses spacepy and astropy to convert time and position arrays
    from one coordinate system to another. It will correct obvious errors in
    the return units, but may not catch all incorrect values.

    INPUTS:
    inTime array:  time in UTC timestamp
    c1 array:  x (in R_earth)*, lon (in deg)
    c2 array:  y (in R_earth)*, lat (in deg)
    c3 array:  z (in R_earth)*, alt (in km), radius (in R_earth)
    inCoord string: case-sensitive string from list:
        'GDZ', 'GEO', 'GSM', 'GSE', 'SM', 'GEI', 'MAG', 'SPH', 'RLL'
        (SpacePy coordinates)
        'teme', 'icrs', 'fk5', 'fk4', 'itrs', 'galactic', 'galactocentric',
        'cirs', 'tete', 'precessedgeocentric', 'geocentricmeanecliptic',
        'geocentrictrueecliptic', 'hcrs', 'barycentricmeanecliptic',
        'heliocentricmeanecliptic', 'barycentrictrueecliptic',
        'heliocentrictrueecliptic', 'heliocentriceclipticiau76',
        'custombarycentricecliptic', 'lsr', 'lsrk', 'lsrd', 'supergalactic',
        'galacticlsr', 'fk4noeterms' (AstroPy coordinates)
        Note: Not compatible with AstroPy's HADec and AltAz coordinate systems.
        Note: Conversions using the galactocentric coordinate system are not
            conserved (a conversion to and then from this coordinate system
                       does not return the same beginning values).
    inType string: car, sph
    outCoord string: (same list as for inCoord string)
    outType string: car, sph

    OUTPUT:
    c1 array:  x (in R_earth)*, lon (in deg)
    c2 array:  y (in R_earth)*, lat (in deg)
    c3 array:  z (in R_earth)*, alt (in km), radius (in R_earth)
    units array:  [unit_c1, unit_c2, unit_c3]  (for example ['deg','deg','km']
                                                    or ['R_E','R_E','R_E'])
    *SpacePy's GDZ car coordinate system requires and produces (x, y, z) in km.

    The resource information on SpacePy's coordinate conversion function is
        sparse at best, so the below information has been collected via other
        resources and our own testing of the function. The data concerning the
        spherical coordinate systems are collected into a table format for
        easier perusal. Some details concerning the AstroPy coordinate systems
        are also below.
    For cartesian coordinates, all of the input values after time should be in
        earth radii (R_E) in order (x, y, z) to work properly, except for GDZ.
    For spherical coordinates, all of the input values after time should be in
        order (longitude, latitude, altitude or radius). The longitude and
        latitude values should be in degrees, altitude values in kilometers,
        and radius values in earth radii (R_E) from the Earth's center. All
        latitude values should fall between -90 and 90 degrees. The longitude
        range differs between the coordinate systems and is given for each in
        the table below.
    The intepretations of the input coordinates vary between the AstroPy
        coordinate systems (see below). We leave it to the user to determine
        the proper input values accordingly.
    The longitude values returned for a given coordinate converted to an
        AstroPy coordinate system are always positive (0 to 360 degrees).

    SpacePy
    Abbrev.   Full Name                       Lon. range     vertical variable
    --------------------------------------------------------------------------
    GDZ    Geodetic (WGS 84)                  (-180, 180)    Altitude (km)
    GEO    Geographic                         (-180, 180)    Radius (R_E)
    GSM    Geocentric Solar Magnetospheric    (-180, 180)    Radius (R_E)
    GSE    Geocentric Solar Ecliptic          (-180, 180)    Radius (R_E)
    SM     Solar Magnetic                     (-180, 180)    Radius (R_E)
    GEI    Geocentric Equatorial Inertial     (-180, 180)    Radius (R_E)
          (also ECI = Earth-Centered Inertial)
    MAG    Geomagnetic                        (-180, 180)    Radius (R_E)
    SPH    Spherical                            (0, 360)     Radius (R_E)
    RLL    Radius, Latitude, Longitude        (-180, 180)    Radius (R_E)

    For descriptions of most of the coordinate systems, see
    https://sscweb.gsfc.nasa.gov/users_guide/Appendix_C.shtml and
    "Geophysical Coordinate Transformations", C.T. Russell, Cosmic
        Electrodynamics, Vol. 2, pp. 184 - 196, 1971.
    The current links to SpacePy's coordinate documentation and wrapped
    conversion functions are:
        https://spacepy.github.io/autosummary/spacepy.coordinates.Coords.html
        http://svn.code.sf.net/p/irbem/code/trunk/manual/user_guide.html

    AstroPy coordinate systems:
        https://docs.astropy.org/en/stable/coordinates/skycoord.html
    Spherical coordinates in the AstroPy coordinate systems are interpreted as
    described below with the units [x,y,z] = [deg, deg, R_E].
    (All z values must be in R_E, not km.)
        teme: ['lon', 'lat', 'distance']
        icrs: ['ra', 'dec', 'distance']
        fk5: ['ra', 'dec', 'distance']
        fk4: ['ra', 'dec', 'distance']
        itrs: ['lon', 'lat', 'distance']
        galactic: ['l', 'b', 'distance']
        galactocentric: ['lon', 'lat', 'distance']
        cirs: ['ra', 'dec', 'distance']
        tete: ['ra', 'dec', 'distance']
        precessedgeocentric: ['ra', 'dec', 'distance']
        geocentricmeanecliptic: ['lon', 'lat', 'distance']
        geocentrictrueecliptic: ['lon', 'lat', 'distance']
        hcrs: ['ra', 'dec', 'distance']
        barycentricmeanecliptic: ['lon', 'lat', 'distance']
        heliocentricmeanecliptic: ['lon', 'lat', 'distance']
        barycentrictrueecliptic: ['lon', 'lat', 'distance']
        heliocentrictrueecliptic: ['lon', 'lat', 'distance']
        heliocentriceclipticiau76: ['lon', 'lat', 'distance']
        custombarycentricecliptic: ['lon', 'lat', 'distance']
        lsr: ['ra', 'dec', 'distance']
        lsrk: ['ra', 'dec', 'distance']
        lsrd: ['ra', 'dec', 'distance']
        galacticlsr: ['l', 'b', 'distance']
        fk4noeterms: ['ra', 'dec', 'distance']
        supergalactic: ['sgl', 'sgb', 'distance']
    Cartesian coordinates in the AstroPy coordinate systems are interpreted as
    (x,y,z) with a few exceptions (below) and always in R_E (earth radii).
        galactic: ['u', 'v', 'w']
        supergalactic: ['sgx', 'sgy', 'sgz']
    """

    # Start timer, define coordinate system lists in each package
    tic = time.perf_counter()
    if (inCoord not in astropy_coordlist) and (inCoord not in
                                               spacepy_coordlist):
        raise ValueError(f'The {inCoord} coordinate system is not recogized.' +
                         f'Must be one of: {spacepy_coordlist} ' +
                         f'{astropy_coordlist}.')

    # convert input values to arrays with a defined length to avoid errors.
    inTime = convert_to_array(inTime)
    c1 = convert_to_array(c1)
    c2 = convert_to_array(c2)
    c3 = convert_to_array(c3)

    # skip conversions to same coordinate system and simply return
    if inCoord == outCoord and inType == outType:
        # figure out units to return
        if verbose:
            print(f'No conversion necessary from {inCoord} {inType} to ' +
                  f'{outCoord} {outType}. Returning. ', end="")
        if inType == 'sph':
            units = ['deg', 'deg', 'R_E']
        if inType == 'sph' and inCoord == 'GDZ':
            units[2] = 'km'
        if inType == 'car':
            units = ['R_E', 'R_E', 'R_E']
        toc = time.perf_counter()
        if verbose:
            print(f'Elapsed time: {toc-tic:.4f} seconds.')
        return c1, c2, c3, units
    if len(c1) > 10000 or verbose:
        print(f'Converting {len(c1)} positions into {outCoord+outType} ' +
              'coordinates...', end="")
    # Create coordinate object
    if inCoord in spacepy_coordlist:
        coord_obj = create_spacepy(inTime, c1, c2, c3, inCoord, inType)
    elif inCoord in astropy_coordlist:
        coord_obj = create_astropy(inTime, c1, c2, c3, inCoord, inType)

    # Perform coordinate conversion
    if inCoord in spacepy_coordlist and outCoord in spacepy_coordlist:
        new_coord = spacepy_spacepy(coord_obj, outCoord, outType)
    elif inCoord in astropy_coordlist and outCoord in astropy_coordlist:
        new_coord = astropy_astropy(coord_obj, outCoord, outType)
    elif inCoord in spacepy_coordlist and outCoord in astropy_coordlist:
        new_coord = spacepy_astropy(coord_obj, outCoord, outType)
    elif inCoord in astropy_coordlist and outCoord in spacepy_coordlist:
        new_coord = astropy_spacepy(coord_obj, outCoord, outType)

    # Extract x, y, z, and units from converted coordinate object
    if outCoord in spacepy_coordlist:
        x, y, z, units = extract_spacepy(inTime, c1, c2, c3, inCoord, inType,
                                         outCoord, outType, new_coord,
                                         verbose=verbose)
    elif outCoord in astropy_coordlist:
        x, y, z, units = extract_astropy(new_coord, outType)

    # Ending messages and final return.
    toc = time.perf_counter()
    if len(c1) > 10000 or verbose:
        print(f'done in {toc-tic:0.4f} seconds.')
    return x, y, z, units


def ComputeLshell(inTime, c1, c2, c3, inCoord, inType):
    """
    This function uses spacepy to compute L shell and MLT from given time and
    position values. It uses the IRBEM library to do the calculations.
    (Not currently used in Kamodo.)

    INPUTS:
    inTime:       array:  time in UTC timestamp
    c1:           array:  x(R_E)*, lon(deg)
    c2:           array:  y(R_E)*, lat(deg)
    c3:           array:  z(R_E)*, alt(km), radius(R_E)
    *SpacePy's GDZ car coordinate system requires (x, y, z) in km.
    inCoordName:  string: case-sensitive string from list in ConvertCood. See
        that function's documentation for more details.

    OUTPUT:
    c1:       array:  Lm
    c2:       array:  Lstar
    c3:       array:  MLT
    """

    from spacepy.irbempy import get_Lstar

    # Start timer
    tic = time.perf_counter()

    # convert input values to arrays with a defined length to avoid errors
    inTime = convert_to_array(inTime)
    c1 = convert_to_array(c1)
    c2 = convert_to_array(c2)
    c3 = convert_to_array(c3)

    # create Spacepy Coord, converting from astropy coord sys if needed
    if inCoord in spacepy_coordlist:
        spacepy_coord = create_spacepy(inTime, c1, c2, c3, inCoord, inType)
        cvals = spacepy_spacepy(spacepy_coord, 'GSM', 'sph')
    elif inCoord in astropy_coordlist:
        astropy_coord = create_astropy(inTime, c1, c2, c3, inCoord, inType)
        cvals = astropy_spacepy(astropy_coord, 'GSM', 'sph')

    # compute Lstar
    # Need to test what spacepy coord sys this works fastest in
    # (GSM sph as above?)
    qq = get_Lstar(cvals.ticks, cvals)
    Lm = abs(np.reshape(qq['Lm'], -1))
    Lstar = abs(np.reshape(qq['Lstar'], -1))
    MLT = qq['MLT']

    # Ending statements and return
    toc = time.perf_counter()
    print('Computed Lm, lstar, MLT in ', "{:0.4f}".format(toc-tic),
          ' seconds for', len(inTime), ' positions.')
    return Lm, Lstar, MLT


# Not sure we need this. If so, will need to add integers for astropy
# coordinates.
def ConvertCoord_f(inTime, c1, c2, c3, iC1, iT1, iC2, iT2):
    """
    Wrapper function to ConvertCoord for Fortran calls.

    iC1,iC2 mapping:
        0 = GDZ   3 = GSE   6 = MAG
        1 = GEO   4 = SM    7 = SPH
        2 = GSM   5 = GEI   8 = RLL

    iT1,iT2 mapping:
        0 = car   1 = sph
    """

    Coords = ['GDZ', 'GEO', 'GSM', 'GSE', 'SM', 'GEI', 'MAG', 'SPH', 'RLL']
    Types = ['car', 'sph']

    inCoord = Coords[iC1]
    outCoord = Coords[iC2]
    inType = Types[iT1]
    outType = Types[iT2]

    x, y, z, units = ConvertCoord(inTime, c1, c2, c3, inCoord, inType,
                                  outCoord, outType)
    return x, y, z



# -*- coding: utf-8 -*-
"""
Created on Thu Jun  3 17:11:29 2021

@author: Darren
Adapted by R Ringuette
"""

from kamodo import Kamodo, kamodofy
import numpy as np
import datetime
import urllib, json
# Renamed hapi as PYhapi to avoid name conflict
from hapiclient import hapi as PYhapi
from hapiclient import hapitime2datetime
import plotly.express as px
import plotly.graph_objects as go
from astropy.constants import R_earth
import pandas as pd
from pandas import DatetimeIndex
from collections.abc import Iterable


'''Examples:
server     = 'http://hapi-server.org/servers/SSCWeb/hapi'
dataset    = 'grace1'
parameters = 'X_GEO,Y_GEO,Z_GEO,X_GSE,Y_GSE,Z_GSE'
start      = '2012-07-07T00:00:00'
stop       = '2012-07-08T00:00:00'
hapi = HAPI(server, dataset, parameters, start, stop)

server2 = 'https://cdaweb.gsfc.nasa.gov/hapi'
dataset2 = 'GOES12_K0_MAG'
parameters2 = 'B_GSE_c,SC_pos_eo,SC_pos_se'
start2      = '2008-07-11T00:00:00'
stop2       = '2008-07-13T00:00:00'
hapi2 = HAPI(server2, dataset2, parameters2, start2, stop2)
'''

def hapi_get_parameters(server, dataset):
    '''return only a list of the possible parameters'''
    data = hapi_get_info(server, dataset)
    return [data['parameters'][i]['name'] for i in range(len(data['parameters']))]

def hapi_get_info(server, dataset):
    '''Query for info json return from HAPI server'''
    query = '{}/info?id={}'.format(server, dataset)
    response = urllib.request.urlopen(query)
    data = json.loads(response.read())
    return data

def hapi_get_date_range(server, dataset):
    '''Query for start and end date for dataset'''
    info = hapi_get_info(server, dataset)
    start_date = info['startDate']
    end_date = info['stopDate']
    return start_date, end_date

class HAPI(Kamodo):
    def __init__(self, server, dataset, parameters = None, start = None, stop = None,
                register_components=True,
                 verbose=False, **kwargs):
        super(HAPI, self).__init__(**kwargs)
        self.verbose=verbose
        self.symbol_registry=dict()
        self.signatures=dict()
        self.RE=R_earth.value
        self.server = server
        self.dataset = dataset
        if verbose: print(' -server: ',server)
        if verbose: print(' -dataset: ',dataset)
        opts       = {'logging': False, 'usecache': False}
        if start is None or stop is None:
            start, stop = hapi_get_date_range(server, dataset)
            # Some HAPI stop strings have fractional seconds and some don't, look for decimal.
            if "." in stop:
                ts = datetime.datetime.strptime(stop, '%Y-%m-%dT%H:%M:%S.%fZ').timestamp()
            else:
                ts = datetime.datetime.strptime(stop, '%Y-%m-%dT%H:%M:%SZ').timestamp()
            # Make sure stop format same as start
            stop  = datetime.datetime.fromtimestamp(ts).strftime("%Y-%m-%dT%H:%M:%S.%fZ")
            ts = ts - 3600.
            start = datetime.datetime.fromtimestamp(ts).strftime("%Y-%m-%dT%H:%M:%S.%fZ")
            print(" -dates not selected, using last hour available.")
            print("       start: ",start)
            print("         end: ",stop)
        else:
            if verbose: print(" -date start: ",start)
            if verbose: print("         end: ",stop)
        self.start=start
        self.stop=stop
        # Get data, meta from the HAPI python client
        data, meta = PYhapi(server, dataset, parameters, start, stop, **opts)
        self.hdata = data
        self.meta = meta
        self.startDate = meta['startDate']
        self.stopDate = meta['stopDate']
        # Convert binary time array into datetime
        self.dtarray = pd.to_datetime(hapitime2datetime(data['Time']))
        self.tsarray = np.array([d.timestamp() for d in self.dtarray])
        self.variables=dict()
        for metaparam in self.meta['parameters']:
            varname = metaparam['name']
            if varname != "Time":
                # Check for dimentionality of variable (scalar=1, vector = 3, others?)
                asize = 1
                if "size" in metaparam:
                    asize = metaparam['size'][0]
                aunit = metaparam['units']
                adata = self.hdata[varname]
                afill = metaparam['fill']
                adesc = metaparam['description']
                if "0.1nT" in aunit:
                    # Fix wonky geotail data
                    aunit="nT"
                    afill = 0.1*float(afill)
                    adata = 0.1*adata.astype(np.float)
                if aunit == "n/cc":
                    # The unit n/cc should be 1/cc to register properly in Kamodo
                    aunit = "1/cc"
                if aunit == " ":
                    # A space is not valid for units
                    aunit = ""
                self.variables[varname] = dict(units=aunit,
                                               data=adata,
                                               size=asize,
                                               fill=afill,
                                               desc=adesc)
        for varname in self.variables:
            units = self.variables[varname]['units']
            if verbose: print('... registering ',varname,units)
            self.register_variable(varname, units)
        
        # classification of position into coordinates to assist visualizion
        self.possible_coords=('TOD','J2K','GEO','GM','GSM','GSE','SM')
        self.possible_directions=('x','y','z')
        self.coords=dict()
        for varname in self.variables:
            if '_' not in varname:
                continue
            size = self.variables[varname]['size']
            desc = self.variables[varname]['desc']
            if size == 1:
                # Look for position values from SSCWeb, ie. X_GSE, Y_GSE, Z_GSE
                tmp = varname.split("_")
                direction = tmp[0].lower()
                key = tmp[1]
                if key in self.possible_coords and direction in self.possible_directions:
                    if key not in self.coords:
                        self.coords[key] = dict(coord=key)
                        if verbose: print("... position vector registered in",key)
                    self.coords[key]['size'] = size
                    self.coords[key][direction] = varname
            elif size ==3:
                # Look for position or vector values, ie. SC_pos_se, POS
                tmp = varname.split("_")
                if tmp[0].lower() != "pos":
                    if tmp[1].lower() != "pos":
                        continue
                key = "none"
                if len(tmp) > 2:
                    if tmp[2].lower() == "se":
                        key = "GSE"
                    elif tmp[2].lower() == "sm":
                        key = "GSM"
                    elif tmp[2].lower() == "eo":
                        key = "GEO"
                if len(tmp) == 1:
                    if "coordinate" in desc:
                        if "GSE" in desc:
                            key = "GSE"
                        elif "GSM" in desc:
                            key = "GSM"
                        elif "GEO" in desc:
                            key = "GEO"
                if key in self.possible_coords:
                    if key not in self.coords:
                        self.coords[key] = dict(coord=key)
                        if verbose: print("... position vector registered in",key)
                    self.coords[key]['size'] = size
                    self.coords[key]['x'] = varname
                    self.coords[key]['y'] = varname
                    self.coords[key]['z'] = varname

        # Change 'fill' values in data to NaNs
        self.fill2nan(verbose=verbose)
        
        if register_components:
            self.register_components()

    def register_variable(self, varname, units):
        """register variables into Kamodo for this service, HAPI"""
        data =  self.variables[varname]['data']
        times = self.dtarray

        if self.verbose:
            print('{} shape {}'.format(varname, data.shape))

        isvector = False
        try:
            if len(data.shape) > 1:
                isvector = True
                ser = pd.DataFrame(data, index=pd.DatetimeIndex(times))
            else:
                ser = pd.Series(data, index=pd.DatetimeIndex(times))
        except:
            raise NotImplementedError('data issue: {}'.format(data.shape))

        def interpolate(t=times):
            ts = t
            isiterable = isinstance(t, Iterable)

            if isinstance(ts, DatetimeIndex):
                pass
            elif isinstance(ts, float):
                ts = pd.to_datetime([ts], utc=True, unit='s')
            elif isiterable:
                if isinstance(ts[0], float):
                    ts = pd.to_datetime(ts, utc=True, unit='s')
                ts = DatetimeIndex(ts)
            else:
                raise NotImplementedError(ts)
            
            ser_ = ser.reindex(ser.index.union(ts))
            ser_interpolated = ser_.interpolate(method='time')
            result = ser_interpolated.reindex(ts)
            if isiterable:
                return result.values
            else:
                return result.values[0]

        # store the interpolator
        self.variables[varname]['interpolator'] = interpolate

        # update docstring for this variable
        for _ in self.meta['parameters']:
            if _['name'] == 'B_GSE_c':
                interpolate.__doc__ = "{} in [{}].".format(_['description'], units)

        self[varname] = kamodofy(interpolate, 
                                 units = units, 
                                 citation = "De Zeeuw 2020",
                                 data = None)


    def register_hat(self, component_id, direction):
        @kamodofy
        def component(xvec):
            return xvec[:, component_id]

        self['{}hat'.format(direction)] = component

    def register_components(self):
        for i, _ in enumerate(self.possible_directions):
            self.register_hat(i, _)

        for param in self.meta['parameters']:
            if param.get('size', [1])[-1] > 1:
                var_orig = param['name']
                var_split = var_orig.split('_')
                var_base = var_split[0]
                var_end = '_'.join(var_split[1:])
                for component in self.possible_directions:
                    var_hat = '{}hat'.format(component)
                    var_name = '{}_{}__{}'.format(var_base, component, var_end)
                    self[var_name] = '{}({})'.format(var_hat, var_orig)

    def fill2nan(self, verbose=False):
        '''
        Replaces fill value in data with NaN.
        '''
        for varname in self.variables:
            data = self.variables[varname]['data']
            fill = self.variables[varname]['fill']
            if fill != None and fill != "NaN":
                # NOTE: If values are integers and nbad > 0, it will throw errer.
                mask = data==float(fill)
                nbad = np.count_nonzero(mask)
                if nbad > 0:
                    if verbose: print("Found",nbad,"fill values, replacing with NaN for variable",
                          varname,"of size",data.size)
                    data[mask]=np.nan
                self.variables[varname]['data'] = data
        
    def get_plot(self, coord="", type="1Dpos", scale="R_E", var=""):
        '''
        Return a plotly figure object in a selected coordinate system.
        coord = coordinate system for position plots
        type = 1Dvar => 1D plot of variable value vs Time
               1Dpos (default) => 1D location x,y,z vs Time
               3Dpos => 3D location colored by altitude
        scale = km, R_E (default)
        var = variable name for variable value plots
        '''

        # Set plot title for first plots
        txttop=self.dataset + " data from " + self.server + "<br>"\
            + self.start + " to " + self.stop

        if type == '1Dvar':
            if var == "":
                print("No plot variable passed in.")
                return
            fig=go.Figure()
            if self.variables[var]['size'] == 1:
                x=self.variables[var]['data']
                fig.add_trace(go.Scatter(x=self.dtarray, y=x, mode='lines+markers', name=var))
            elif self.variables[var]['size'] == 3:
                x=self.variables[var]['data'][:,0]
                y=self.variables[var]['data'][:,1]
                z=self.variables[var]['data'][:,2]
                fig.add_trace(go.Scatter(x=self.dtarray, y=x, mode='lines+markers', name=var))
                fig.add_trace(go.Scatter(x=self.dtarray, y=y, mode='lines+markers', name=var))
                fig.add_trace(go.Scatter(x=self.dtarray, y=z, mode='lines+markers', name=var))
            ytitle=var+" ["+self.variables[var]['units']+"]"
            fig.update_xaxes(title_text="Time")
            fig.update_yaxes(title_text=ytitle)
            fig.update_layout(hovermode="x")
            fig.update_layout(title_text=txttop)
 
            return fig
        
        # Set plot title for next plots
        txttop=self.dataset + " data from " + self.server + "<br>"\
            + self.start + " to " + self.stop + "<br>" + coord
        
        if coord not in self.coords:
            print('ERROR, sent in coords not in dataset.')
            return
        
        if type == '1Dpos':
            fig=go.Figure()
            xvarname = self.coords[coord]['x']
            if self.coords[coord]['size'] == 1:
                x=self.variables[self.coords[coord]['x']]['data']
                y=self.variables[self.coords[coord]['y']]['data']
                z=self.variables[self.coords[coord]['z']]['data']
            elif self.coords[coord]['size'] == 3:
                x=self.variables[self.coords[coord]['x']]['data'][:,0]
                y=self.variables[self.coords[coord]['y']]['data'][:,1]
                z=self.variables[self.coords[coord]['z']]['data'][:,2]
            if scale == "km":
                if self.variables[xvarname]['units'] == "R_E":
                    x=x*self.RE
                    y=y*self.RE
                    z=z*self.RE
                ytitle="Position [km]"
            else:
                if self.variables[xvarname]['units'] == "km":
                    x=x/self.RE
                    y=y/self.RE
                    z=z/self.RE
                ytitle="Position [R_E]"
            fig.add_trace(go.Scatter(x=self.dtarray, y=x,
                                     mode='lines+markers', name=self.coords[coord]['x']))
            fig.add_trace(go.Scatter(x=self.dtarray, y=y,
                                     mode='lines+markers', name=self.coords[coord]['y']))
            fig.add_trace(go.Scatter(x=self.dtarray, y=z,
                                     mode='lines+markers', name=self.coords[coord]['z']))
            fig.update_xaxes(title_text="Time")
            fig.update_yaxes(title_text=ytitle)
            fig.update_layout(hovermode="x")
            fig.update_layout(title_text=txttop)
 
            return fig
        
        if type == "3Dpos":
            xvarname = self.coords[coord]['x']
            if self.coords[coord]['size'] == 1:
                x=self.variables[self.coords[coord]['x']]['data']
                y=self.variables[self.coords[coord]['y']]['data']
                z=self.variables[self.coords[coord]['z']]['data']
            elif self.coords[coord]['size'] == 3:
                x=self.variables[self.coords[coord]['x']]['data'][:,0]
                y=self.variables[self.coords[coord]['y']]['data'][:,1]
                z=self.variables[self.coords[coord]['z']]['data'][:,2]
            if scale == "km":
                if self.variables[xvarname]['units'] == "R_E":
                    x=x*self.RE
                    y=y*self.RE
                    z=z*self.RE
                r=(np.sqrt(x**2 + y**2 + z**2))-self.RE
                ytitle="Position [km]"
            else:
                if self.variables[xvarname]['units'] == "km":
                    x=x/self.RE
                    y=y/self.RE
                    z=z/self.RE
                r=(np.sqrt(x**2 + y**2 + z**2))-1.
                ytitle="Position [R_E]"
            fig=px.scatter_3d(
                x=x,
                y=y,
                z=z,
                color=r)
            bartitle = "Altitude [" + scale + "]"
            fig.update_layout(coloraxis=dict(colorbar=dict(title=bartitle)))
            fig.update_layout(scene=dict(xaxis=dict(title=dict(text="X ["+scale+"]")),
                                         yaxis=dict(title=dict(text="Y ["+scale+"]")),
                                         zaxis=dict(title=dict(text="Z ["+scale+"]"))))
            fig.update_layout(title_text=txttop)
            return fig
        

        print('ERROR, reached end of get_plot without any action taken.')
        return



'''
Original version: Lutz Raestatter Oct 1(?), 2021
Modify to work with flythrough: Oct 5, 2021 (Rebecca Ringuette)
'''
from datetime import datetime, timezone

# standard model dictionary for reference
model_varnames = {'bx': ['B_x', 'x component of magnetic field',
                         0, 'GSE', 'car', ['time', 'x', 'y', 'z'], 'nT'],
                  'by': ['B_y', 'y component of magnetic field',
                         1, 'GSE', 'car', ['time', 'x', 'y', 'z'], 'nT'],
                  'bz': ['B_z', 'z component of magnetic field',
                         2, 'GSE', 'car', ['time', 'x', 'y', 'z'], 'nT'],
                  'ex': ['E_x', 'x component of electric field',
                         6, 'GSE', 'car', ['time', 'x', 'y', 'z'], 'mV/m'],
                  'ey': ['E_y', 'y component of electric field',
                         7, 'GSE', 'car', ['time', 'x', 'y', 'z'], 'mV/m'],
                  'ez': ['E_z', 'z component of electric field',
                         8, 'GSE', 'car', ['time', 'x', 'y', 'z'], 'mV/m'],
                  'vx': ['v_plasmax', 'x component of plasma velocity',
                         9, 'GSE', 'car', ['time', 'x', 'y', 'z'], 'km/s'],
                  'vy': ['v_plasmay', 'y component of plasma velocity',
                         10, 'GSE', 'car', ['time', 'x', 'y', 'z'], 'km/s'],
                  'vz': ['v_plasmaz', 'z component of plasma velocity',
                         11, 'GSE', 'car', ['time', 'x', 'y', 'z'], 'km/s'],
                  'rr': ['N_plasma', 'number density of plasma ' +
                         '(hydrogen equivalent)',
                         12, 'GSE', 'car', ['time', 'x', 'y', 'z'], '1/cm**3'],
                  'resis': ['eta', 'resistivity',
                            13, 'GSE', 'car', ['time', 'x', 'y', 'z'],
                            'm**2/s'],
                  'pp': ['P_plasma', 'plasma pressure',
                         14, 'GSE', 'car', ['time', 'x', 'y', 'z'], 'pPa'],
                  'xjx': ['j_x', 'current density, x component',
                          15, 'GSE', 'car', ['time', 'x', 'y', 'z'],
                          'muA/m**2'],
                  'xjy': ['j_y', 'current density, y component',
                          16, 'GSE', 'car', ['time', 'x', 'y', 'z'],
                          'muA/m**2'],
                  'xjz': ['j_z', 'current density, z component',
                          17, 'GSE', 'car', ['time', 'x', 'y', 'z'],
                          'muA/m**2']}


def MODEL():
    from numpy import array, unique, squeeze
    from time import perf_counter
    from os.path import isfile, basename
    from glob import glob
    from netCDF4 import Dataset
    from kamodo import Kamodo
    import kamodo_ccmc.readers.reader_utilities as RU

    class MODEL(Kamodo):
        '''OpenGGCM_GM magnetosphere reader.

        Inputs:
            file_dir: a string representing the file directory of the
                model output data.
                Note: This reader 'walks' the entire dataset in the directory.
            variables_requested = a list of variable name strings chosen from
                the model_varnames dictionary in this script, specifically the
                first item in the list associated with a given key.
                - If empty, the reader functionalizes all possible variables
                    (default)
                - If 'all', the reader returns the model_varnames dictionary
                    above for only the variables present in the given files.
            filetime = boolean (default = False)
                - If False, the script fully executes.
                - If True, the script only executes far enough to determine the
                    time values associated with the chosen data.
            printfiles = boolean (default = False)
                - If False, the filenames associated with the data retrieved
                    ARE NOT printed.
                - If True, the filenames associated with the data retrieved ARE
                    printed.
            gridded_int = boolean (default = True)
                - If True, the variables chosen are functionalized in both the
                    standard method and a gridded method.
                - If False, the variables chosen are functionalized in only the
                    standard method.
            verbose = boolean (False)
                - If False, script execution and the underlying Kamodo
                    execution is quiet except for specified messages.
                - If True, be prepared for a plethora of messages.
        All inputs are described in further detail in
            KamodoOnboardingInstructions.pdf.

        Returns: a kamodo object (see Kamodo core documentation) containing all
            requested variables in functionalized form.

        Notes and special features:
            - The file converter for the OpenGGCM global magnetosphere outputs
              (compressed binary files) currently only runs on CCMC machines.
              Please contact CCMC for the desired run to be converted to
              netCDF4 (Lutz Rastaetter).
            - This model reader has two special properties called
              kamodo_object.near_Earth_boundary_radius and
              kamodo_object.near_Earth_boundary_radius_unit that give the
              inner boundaries of the radial domain for the given run. The
              inner boundary will also be readily apparent when viewing any
              plot including the coordinate origin (X, Y, Z) = (0, 0, 0). The
              unit of the inner boundary is typically earth radii (R_E).
            - The model outputs are produced with one time step per file, so
              interpolation method 1 is chosen. The standard SciPy interpolator
              is used.
                
        '''
        def __init__(self, file_dir, variables_requested=[],
                     filetime=False, verbose=False, gridded_int=True,
                     printfiles=False, **kwargs):
            super(MODEL, self).__init__()
            self.modelname = 'OpenGGCM_GM'
            t0 = perf_counter()  # profiling time stamp

            # first, check for file list, create if DNE
            list_file = file_dir + self.modelname + '_list.txt'
            time_file = file_dir + self.modelname + '_times.txt'
            self.times, self.pattern_files = {}, {}
            if not isfile(list_file) or not isfile(time_file):
                # collect filenames
                files = sorted(glob(file_dir+'*.nc'))
                if len(files) == 0:
                    try:
                        from kamodo_ccmc.readers.openggcm_to_cdf import \
                            openggcm_combine_magnetosphere_files as gmconv
                        self.conversion_test = gmconv(file_dir)
                    except:
                        print('The file converter for the OpenGGCM global ' +
                              'magnetosphere outputs currently only runs on ' +
                              'CCMC machines. Please contact CCMC for the ' +
                              'files for the desired run converted to ' +
                              'netCDF4 from Lutz Rastaetter.')
                        return
                patterns = unique([basename(file[:-19]) for file in files])
                self.filename = ''.join([f+',' for f in files])[:-1]
                self.filedate = datetime.strptime(
                    files[0][-19:-9]+' 00:00:00', '%Y-%m-%d %H:%M:%S'
                    ).replace(tzinfo=timezone.utc)

                # establish time attributes
                for p in patterns:
                    # get list of files to loop through later
                    pattern_files = sorted(glob(file_dir+p+'*.nc'))
                    self.pattern_files[p] = pattern_files
                    self.times[p] = {'start': [], 'end': [], 'all': []}

                    # loop through to get times
                    self.times[p]['start'] = array([RU.str_to_hrs(
                        f[-19:-3]+'_00', self.filedate,
                        format_string='%Y-%m-%d_%H_%M_%S') for f in
                        pattern_files])
                    self.times[p]['end'] = self.times[p]['start'].copy()
                    self.times[p]['all'] = self.times[p]['start'].copy()

                # create time list file if DNE
                RU.create_timelist(list_file, time_file, self.modelname,
                                   self.times, self.pattern_files,
                                   self.filedate)
            else:  # read in data and time grids from file list
                self.times, self.pattern_files, self.filedate, self.filename =\
                    RU.read_timelist(time_file, list_file)
            if filetime:
                return  # return times only

            # perform initial check on variables_requested list
            if len(variables_requested) > 0 and variables_requested != 'all':
                test_list = [value[0] for key, value in model_varnames.items()]
                err_list = [item for item in variables_requested if item not in
                            test_list]
                if len(err_list) > 0:
                    print('Variable name(s) not recognized:', err_list)
                for item in err_list:
                    variables_requested.remove(item)
                if len(variables_requested) == 0:
                    return

            # there is only one pattern for OpenGGCM, so just save the one grid
            p = list(self.pattern_files.keys())[0]
            pattern_files = self.pattern_files[p]
            cdf_data = Dataset(pattern_files[0], 'r')

            # check var_list for variables not possible in this file set
            self.err_list = []
            if len(variables_requested) > 0 and\
                    variables_requested != 'all':
                gvar_list = [key for key in model_varnames.keys()
                             if key in cdf_data.variables.keys() and
                             model_varnames[key][0] in variables_requested]
                if len(gvar_list) != len(variables_requested):
                    err_list = [value[0] for key, value in
                                model_varnames.items()
                                if key not in cdf_data.variables.keys() and
                                value[0] in variables_requested]
                    self.err_list.extend(err_list)  # add to master list
            else:
                gvar_list = [key for key in model_varnames.keys()
                             if key in cdf_data.variables.keys()]

            # store which file these variables came from
            self.varfiles = [model_varnames[key][0] for
                             key in gvar_list]
            self.gvarfiles = gvar_list
            # initialize storage structure
            self.variables = {model_varnames[gvar][0]: {
                'units': model_varnames[gvar][-1], 'data': p} for gvar in
                self.gvarfiles}

            # get coordinate grids
            self.near_Earth_boundary_radius = \
                cdf_data.near_Earth_boundary_radius
            self.near_Earth_boundary_radius_unit = \
                cdf_data.near_Earth_boundary_radius_units
            for grid in ['_x', '_y', '_z']:
                setattr(self, grid, array(cdf_data.variables[grid]))
            cdf_data.close()

            # print message if variables not found
            if len(self.err_list) > 0:
                print('Some requested variables are not available: ',
                      self.err_list)

            # collect all possible variables in set of files and return
            if variables_requested == 'all':
                self.var_dict = {value[0]: value[1:] for key, value in
                                 model_varnames.items() if value[0] in
                                 self.varfiles}
                return

            # option to print files
            if printfiles:
                print(f'{len(self.filename)} Files:')
                files = self.filename.split(',')
                for f in files:
                    print(f)

            # register interpolators for each variable
            t_reg = perf_counter()
            # store original list b/c gridded interpolators change keys list
            varname_list = [key for key in self.variables.keys()]
            for varname in varname_list:
                self.register_variable(varname, gridded_int)

            if verbose:
                print(f'Took {perf_counter()-t_reg:.5f}s to register ' +
                      f'{len(varname_list)} variables.')
            if verbose:
                print(f'Took a total of {perf_counter()-t0:.5f}s to kamodofy' +
                      f' {len(varname_list)} variables.')

        # define and register a 4D variable (all are 4D)
        def register_variable(self, varname, gridded_int):
            '''Register and functionalize the variable data.'''

            # determine which file the variable came from, retrieve the coords
            key = self.variables[varname]['data']
            gvar = [key for key, value in model_varnames.items() if
                    value[0] == varname][0]  # variable name in file
            coord_dict = {'time': {'units': 'hr',
                                   'data': self.times[key]['all']}}
            coord_dict['X'] = {'data': self._x, 'units': 'R_E'}
            coord_dict['Y'] = {'data': self._y, 'units': 'R_E'}
            coord_dict['Z'] = {'data': self._z, 'units': 'R_E'}
            coord_str = [value[3]+value[4] for key, value in
                         model_varnames.items() if value[0] == varname][0]

            def func(i):
                '''i is the file/time number. OpenGGCM-mag is one file per
                timestep.'''
                # get data from file
                file = self.pattern_files[key][i]
                cdf_data = Dataset(file)
                data = array(cdf_data.variables[gvar])
                cdf_data.close()
                return squeeze(data)

            # define and register the interpolators
            self = RU.Functionalize_Dataset(self, coord_dict, varname,
                                            self.variables[varname],
                                            gridded_int, coord_str,
                                            interp_flag=1, func=func)
            return
    return MODEL



# -*- coding: utf-8 -*-
"""
Created on Thu Oct  6 15:05:37 2022
@author: rringuet
Data is already in netCDF4, so just doing a few special things here.
The pressure level variables will need to be inverted, so using height in both
    the neutral and density files to calculate
    the median km value for each pressure level across the range of times,
    latitudes and longitudes.
Adapted from the waccmx post processing script.
"""
from numpy import array, median
from time import perf_counter
from netCDF4 import Dataset


def convert_all(file_dir, pattern_files, times):
    '''Find files with height and prepare km post-processing file.'''

    t0 = perf_counter()
    patterns = list(pattern_files.keys())  # e.g. 'height', 'density', 'neutral
    out_file = file_dir + 'CTIPe_km.nc'
    data_out = Dataset(out_file, 'w')
    # loop through neutral/density files and add median/max/min vals to outfile
    for i in range(len(pattern_files[patterns[0]])):
        if 'density' in patterns:  # H_ilev1 logic
            d_data = Dataset(pattern_files['density'][i])
            tmp = d_data.variables['plev']  # plev grid is constant in time
            d_ht = d_data.variables['height']
            if i == 0:
                new_dim = data_out.createDimension(
                    'time1', len(times['density']['start']))
                new_var = data_out.createVariable(
                    'time1', d_ht.datatype, ('time1'))
                new_var[:] = times['density']['start']
                new_dim = data_out.createDimension('ilev1', tmp.shape[0])
                new_var1_t = data_out.createVariable(
                    'km_ilev1_t', d_ht.datatype, tuple(['time1', 'ilev1']))
                new_var1_max = data_out.createVariable(
                    'km_ilev1_max', d_ht.datatype, tuple(['time1']))
                new_var1_min = data_out.createVariable(
                    'km_ilev1_min', d_ht.datatype, tuple(['time1']))
            d_height = array(d_ht)
            new_var1_t[i] = median(d_height, axis=[0, 2, 3])/1000.  # plev = 1
            new_var1_max[i] = d_height.max()/1000.
            new_var1_min[i] = d_height.min()/1000.
            d_data.close()
        if 'neutral' in patterns:  # H_ilev logic
            n_data = Dataset(pattern_files['neutral'][i])
            tmp = n_data.variables['plev']  # plev grid is constant in time
            n_ht = n_data.variables['height']
            if i == 0:
                new_dim = data_out.createDimension(
                    'time', len(times['neutral']['start']))
                new_var = data_out.createVariable(
                    'time', n_ht.datatype, ('time'))
                new_var[:] = times['neutral']['start']
                new_dim = data_out.createDimension('ilev', tmp.shape[0])
                new_var_t = data_out.createVariable(
                    'km_ilev_t', n_ht.datatype, tuple(['time', 'ilev']))
                new_var_max = data_out.createVariable(
                    'km_ilev_max', n_ht.datatype, tuple(['time']))
                new_var_min = data_out.createVariable(
                    'km_ilev_min', n_ht.datatype, tuple(['time']))
            n_height = array(n_ht)
            new_var_t[i] = median(n_height, axis=[0, 2, 3])/1000.  # plev = 1
            new_var_max[i] = n_height.max()/1000.
            new_var_min[i] = n_height.min()/1000.
            n_data.close()
    # calculate median/max/min over time
    if 'density' in patterns:
        new_var = data_out.createVariable('km_ilev1', new_var1_t.datatype,
                                          tuple(['ilev1']))
        new_var[:] = median(array(new_var1_t), axis=0)
        data_out.km_ilev1_max = array(new_var1_max).max()
        data_out.km_ilev1_min = array(new_var1_min).min()
    if 'neutral' in patterns:
        new_var = data_out.createVariable('km_ilev', new_var_t.datatype,
                                          tuple(['ilev']))
        new_var[:] = median(array(new_var_t), axis=0)
        data_out.km_ilev_max = array(new_var_max).max()
        data_out.km_ilev_min = array(new_var_min).min()
    # all done. close output file and return.
    data_out.close()
    print(f'{out_file} created in {perf_counter()-t0}s.')
    return None



# -*- coding: utf-8 -*-
"""
Created on Wed Nov  2 12:59:08 2022

@author: rringuet
DTM model reader
"""

model_varnames = {'Temp_exo': ['T_exo', 'Exospheric temperature', 0, 'GDZ',
                               'sph', ['time', 'lon', 'lat'], 'K'],
                  'Temp': ['T', 'temperature', 0, 'GDZ', 'sph',
                           ['time', 'lon', 'lat', 'height'], 'K'],
                  'DEN': ['rho', 'Total mass density', 0, 'GDZ', 'sph',
                          ['time', 'lon', 'lat', 'height'], 'g/cm**3'],
                  'MU': ['m_avgmol', 'Mean molecular mass', 0, 'GDZ', 'sph',
                         ['time', 'lon', 'lat', 'height'], 'g'],
                  'H': ['N_H', 'Atomic hydrogen partial density', 0, 'GDZ',
                        'sph', ['time', 'lon', 'lat', 'height'], 'g/cm**3'],
                  'He': ['N_He', 'Atomic helium partial density', 0, 'GDZ',
                         'sph', ['time', 'lon', 'lat', 'height'], 'g/cm**3'],
                  'O': ['N_O', 'Atomic oxygen partial density', 0, 'GDZ',
                        'sph', ['time', 'lon', 'lat', 'height'], 'g/cm**3'],
                  'N2': ['N_N2', 'Molecular nitrogen partial density', 0,
                         'GDZ', 'sph', ['time', 'lon', 'lat', 'height'],
                         'g/cm**3'],
                  'O2': ['N_O2', 'Molecular oxygen partial density', 0, 'GDZ',
                         'sph', ['time', 'lon', 'lat', 'height'], 'g/cm**3']
                  }


def MODEL():
    from time import perf_counter
    from glob import glob
    from os.path import basename, isfile
    from numpy import array, unique, NaN, append, transpose, where
    from datetime import datetime, timezone
    from netCDF4 import Dataset
    from kamodo import Kamodo
    import kamodo_ccmc.readers.reader_utilities as RU

    class MODEL(Kamodo):
        '''DTM model data reader.

        Inputs:
            file_dir: a string representing the file directory of the
                model output data.
                Note: This reader 'walks' the entire dataset in the directory.
            variables_requested = a list of variable name strings chosen from
                the model_varnames dictionary in this script, specifically the
                first item in the list associated with a given key.
                - If empty, the reader functionalizes all possible variables
                    (default)
                - If 'all', the reader returns the model_varnames dictionary
                    above for only the variables present in the given files.
            filetime = boolean (default = False)
                - If False, the script fully executes.
                - If True, the script only executes far enough to determine the
                    time values associated with the chosen data.
            printfiles = boolean (default = False)
                - If False, the filenames associated with the data retrieved
                    ARE NOT printed.
                - If True, the filenames associated with the data retrieved ARE
                    printed.
            gridded_int = boolean (default = True)
                - If True, the variables chosen are functionalized in both the
                    standard method and a gridded method.
                - If False, the variables chosen are functionalized in only the
                    standard method.
            verbose = boolean (False)
                - If False, script execution and the underlying Kamodo
                    execution is quiet except for specified messages.
                - If True, be prepared for a plethora of messages.
        All inputs are described in further detail in
            KamodoOnboardingInstructions.pdf.

        Returns: a kamodo object (see Kamodo core documentation) containing all
            requested variables in functionalized form.

        Notes:
            - This model reader is the most basic example of what is required
              in a model reader. The only 'data wrangling' needed is a simple
              longitude wrapping and numpy array transposition to get the
              coordinate order correct.
            - DTM outputs files are given in one netCDF file per day.
            - The files are small and contain multiple time steps per file, so
              interpolation method 2 is chosen. The standard SciPy interpolator
              is used.
        '''

        def __init__(self, file_dir, variables_requested=[],
                     filetime=False, verbose=False, gridded_int=True,
                     printfiles=False, **kwargs):
            super(MODEL, self).__init__()
            self.modelname = 'DTM'

            # first, check for file list, create if DNE
            list_file = file_dir + self.modelname + '_list.txt'
            time_file = file_dir + self.modelname + '_times.txt'
            self.times, self.pattern_files = {}, {}
            if not isfile(list_file) or not isfile(time_file):
                t0 = perf_counter()  # begin timer
                # figure out types of files present (2DTEC, 3DALL, 3DLST, etc)
                files = sorted(glob(file_dir+'*.nc'))
                patterns = sorted(unique([basename(f)[:-11] for f in
                                          files]))  # cut off date
                self.filename = ''.join([f+',' for f in files])[:-1]
                self.filedate = datetime.strptime(
                    basename(files[0])[-10:-3]+' 00:00:00', '%Y%j %H:%M:%S'
                    ).replace(tzinfo=timezone.utc)

                # establish time attributes
                for p in patterns:  # only one pattern
                    # get list of files to loop through later
                    pattern_files = sorted(glob(file_dir+p+'*.nc'))
                    self.pattern_files[p] = pattern_files
                    self.times[p] = {'start': [], 'end': [], 'all': []}

                    # loop through to get times, one day per file
                    for f in range(len(pattern_files)):
                        cdf_data = Dataset(pattern_files[f])
                        # minutes since 12am EACH file -> hrs since 12am 1st f
                        tmp = array(cdf_data.variables['time'])/60. + f*24.
                        self.times[p]['start'].append(tmp[0])
                        self.times[p]['end'].append(tmp[-1])
                        self.times[p]['all'].extend(tmp)
                        cdf_data.close()
                    self.times[p]['start'] = array(self.times[p]['start'])
                    self.times[p]['end'] = array(self.times[p]['end'])
                    self.times[p]['all'] = array(self.times[p]['all'])

                # create time list file if DNE
                RU.create_timelist(list_file, time_file, self.modelname,
                                   self.times, self.pattern_files,
                                   self.filedate)

            else:  # read in data and time grids from file list
                self.times, self.pattern_files, self.filedate, self.filename =\
                    RU.read_timelist(time_file, list_file)
            if filetime:
                return  # return times as is to prevent infinite recursion

            # store variables
            self.missing_value = NaN
            self.varfiles = {}  # store which variable came from which file
            self.gvarfiles = {}  # store file variable name similarly
            self.err_list = []
            self.variables = {}

            # perform initial check on variables_requested list
            if len(variables_requested) > 0 and variables_requested != 'all':
                test_list = [value[0] for key, value in model_varnames.items()]
                err_list = [item for item in variables_requested if item not in
                            test_list]
                if len(err_list) > 0:
                    print('Variable name(s) not recognized:', err_list)
                for item in err_list:
                    variables_requested.remove(item)
                if len(variables_requested) == 0:
                    return

            # there is only one pattern for DTM, so just save the one grid
            p = list(self.pattern_files.keys())[0]
            pattern_files = self.pattern_files[p]

            # get coordinate grids from first file
            cdf_data = Dataset(pattern_files[0], 'r')
            self._lat = array(cdf_data.variables['lat'])  # -90 to 90
            lon = array(cdf_data.variables['lon'])  # 0 to 360
            lon_le180 = list(where(lon <= 180)[0])  # 0 to 180
            lon_ge180 = list(where((lon >= 180) & (lon < 360.))[0])
            self._lon_idx = lon_ge180 + lon_le180
            self._lon = lon - 180.
            self._height = array(cdf_data.variables['ht'])  # km

            # check var_list for variables not possible in this file set
            if len(variables_requested) > 0 and\
                    variables_requested != 'all':
                gvar_list = [key for key in model_varnames.keys()
                             if key in cdf_data.variables.keys() and
                             model_varnames[key][0] in variables_requested]
                if len(gvar_list) != len(variables_requested):
                    err_list = [value[0] for key, value in
                                model_varnames.items()
                                if key not in cdf_data.variables.keys() and
                                value[0] in variables_requested]
                    self.err_list.extend(err_list)  # add to master list
            else:
                gvar_list = [key for key in model_varnames.keys()
                             if key in cdf_data.variables.keys()]

            # store which file these variables came from
            self.varfiles[p] = [model_varnames[key][0] for
                                key in gvar_list]
            self.gvarfiles[p] = gvar_list
            cdf_data.close()

            # print message if variables not found
            if len(self.err_list) > 0:
                print('Some requested variables are not available: ',
                      self.err_list)

            # collect all possible variables in set of files and return
            if variables_requested == 'all':
                var_list = self.varfiles[p]
                self.var_dict = {value[0]: value[1:] for key, value in
                                 model_varnames.items() if value[0] in
                                 var_list}
                return

            # initialize storage structure
            self.variables = {model_varnames[gvar][0]: {
                'units': model_varnames[gvar][-1], 'data': p} for gvar in
                self.gvarfiles[p]}

            # option to print files
            if printfiles:
                print(f'{len(self.filename)} Files:')
                files = self.filename.split(',')
                for f in files:
                    print(f)

            # register interpolators for each variable
            t_reg = perf_counter()
            # store original list b/c gridded interpolators change keys list
            varname_list = [key for key in self.variables.keys()]
            for varname in varname_list:
                self.register_variable(varname, gridded_int)

            if verbose:
                print(f'Took {perf_counter()-t_reg:.5f}s to register ' +
                      f'{len(varname_list)} variables.')
            if verbose:
                print(f'Took a total of {perf_counter()-t0:.5f}s to kamodofy' +
                      f' {len(varname_list)} variables.')

        def register_variable(self, varname, gridded_int):
            """Registers an interpolator with proper signature"""

            # determine which file the variable came from, retrieve the coords
            key = self.variables[varname]['data']
            gvar = [key for key, value in model_varnames.items() if
                    value[0] == varname][0]  # variable name in file
            coord_list = [value[-2] for key, value in
                          model_varnames.items() if value[0] == varname][0]
            coord_dict = {'time': {'units': 'hr',
                                   'data': self.times[key]['all']}}
            # get the correct coordinates
            coord_dict['lon'] = {'data': self._lon, 'units': 'deg'}
            coord_dict['lat'] = {'data': self._lat, 'units': 'deg'}
            if len(coord_list) == 4:
                coord_dict['height'] = {'data': self._height, 'units': 'km'}
            coord_str = [value[3]+value[4] for key, value in
                         model_varnames.items() if value[0] == varname][0]

            # define operations for each variable when given the key
            def func(i):
                '''key is the file pattern, start_idxs is a list of one or two
                indices matching the file start times in self.start_times[key].
                '''
                # get data from file
                file = self.pattern_files[key][i]
                cdf_data = Dataset(file)
                data = array(cdf_data.variables[gvar])
                if hasattr(cdf_data.variables[gvar][0], 'fill_value'):
                    fill_value = cdf_data.variables[gvar][0].fill_value
                else:
                    fill_value = None
                cdf_data.close()
                # if not the last file, tack on first time from next file
                if file != self.pattern_files[key][-1]:  # interp btwn files
                    next_file = self.pattern_files[key][i+1]
                    cdf_data = Dataset(next_file)
                    data_slice = array(cdf_data.variables[gvar][0])
                    cdf_data.close()
                    data = append(data, [data_slice], axis=0)
                # data wrangling
                if fill_value is not None:  # if defined, replace with NaN
                    data = where(data != fill_value, data, NaN)
                if len(data.shape) == 3:
                    variable = transpose(data, (0, 2, 1))
                elif len(data.shape) == 4:
                    variable = transpose(data, (0, 3, 2, 1))
                return variable[:, self._lon_idx]

            self = RU.Functionalize_Dataset(
                self, coord_dict, varname, self.variables[varname],
                gridded_int, coord_str, interp_flag=2, func=func,
                times_dict=self.times[key])

    return MODEL



# -*- coding: utf-8 -*-
"""
Created on Wed Jan  4 15:37:59 2023

@author: rringuet
"""
from numpy import vectorize
from datetime import datetime, timezone

model_varnames = {'Bx': ['B_x', 'X-component of magnetic field',
                         0, 'SM', 'car', ['time', 'X', 'Y', 'Z'], 'nT'],
                  'By': ['B_y', 'Y-component of magnetic field',
                         0, 'SM', 'car', ['time', 'X', 'Y', 'Z'], 'nT'],
                  'Bz': ['B_z', 'Z-component of magnetic field',
                         0, 'SM', 'car', ['time', 'X', 'Y', 'Z'], 'nT'],
                  'Cs': ['c_s', 'Sound speed of plasma', 0, 'SM', 'car',
                         ['time', 'X', 'Y', 'Z'], 'km/s'],
                  'D': ['N_plasma', 'Plasma number density (M/mp)', 0, 'SM',
                        'car', ['time', 'X', 'Y', 'Z'], '1/cm**3'],
                  'Jx': ['J_x', 'X-component of current density',
                         0, 'SM', 'car', ['time', 'X', 'Y', 'Z'], 'nA/m**2'],
                  'Jy': ['J_y', 'Y-component of current density',
                         0, 'SM', 'car', ['time', 'X', 'Y', 'Z'], 'nA/m**2'],
                  'Jz': ['J_z', 'Z-component of current density',
                         0, 'SM', 'car', ['time', 'X', 'Y', 'Z'], 'nA/m**2'],
                  'P': ['P', 'Pressure', 0, 'SM', 'car',
                        ['time', 'X', 'Y', 'Z'], 'nPa'],
                  'Pb': ['P_mag', 'Magnetic pressure', 0, 'SM', 'car',
                         ['time', 'X', 'Y', 'Z'], 'nPa'],
                  'Vx': ['v_x', 'X-component of velocity',
                         0, 'SM', 'car', ['time', 'X', 'Y', 'Z'], 'km/s'],
                  'Vy': ['v_y', 'Y-component of velocity',
                         0, 'SM', 'car', ['time', 'X', 'Y', 'Z'], 'km/s'],
                  'Vz': ['v_z', 'Z-component of velocity',
                         0, 'SM', 'car', ['time', 'X', 'Y', 'Z'], 'km/s'],
                  # CONSTANTS below this line.
                  'dV': ['dV', 'Simulation cell volume', 0, 'SM', 'car',
                         ['X', 'Y', 'Z'], 'R_E**3']
                  }

constants = ['X', 'Y', 'Z', 'dV']


@vectorize
def timestr_timestamp(time_str):
    '''Converts time string from data file into an UTC timestamp. Cuts off
    millisecond portion of times to prevent errors later.'''
    dt = datetime.fromisoformat(time_str).replace(tzinfo=timezone.utc)
    return dt.timestamp()


def MODEL():
    from kamodo import Kamodo
    from glob import glob
    import h5py
    from netCDF4 import Dataset
    from os.path import isfile, basename
    from astropy.time import Time
    from datetime import datetime, timezone
    from numpy import array, NaN, ndarray, zeros, linspace, repeat
    from time import perf_counter
    import kamodo_ccmc.readers.reader_utilities as RU
    import kamodo_ccmc.readers.gameragm_grids as G

    class MODEL(Kamodo):
        '''GAMERA GM model data reader.

        Inputs:
            file_dir: a string representing the file directory of the
                model output data.
                Note: This reader 'walks' the entire dataset in the directory.
            variables_requested = a list of variable name strings chosen from
                the model_varnames dictionary in this script, specifically the
                first item in the list associated with a given key.
                - If empty, the reader functionalizes all possible variables
                    (default)
                - If 'all', the reader returns the model_varnames dictionary
                    above for only the variables present in the given files.
            filetime = boolean (default = False)
                - If False, the script fully executes.
                - If True, the script only executes far enough to determine the
                    time values associated with the chosen data.
            printfiles = boolean (default = False)
                - If False, the filenames associated with the data retrieved
                    ARE NOT printed.
                - If True, the filenames associated with the data retrieved ARE
                    printed.
            gridded_int = boolean (default = True)
                - If True, the variables chosen are functionalized in both the
                    standard method and a gridded method.
                - If False, the variables chosen are functionalized in only the
                    standard method.
            verbose = boolean (False)
                - If False, script execution and the underlying Kamodo
                    execution is quiet except for specified messages.
                - If True, be prepared for a plethora of messages.
        All inputs are described in further detail in
            KamodoOnboardingInstructions.pdf.

        Returns: a kamodo object (see Kamodo core documentation) containing all
            requested variables in functionalized form.

        Notes and instructions:
        - The GAMERA global magnetosphere outputs are given in one or more h5
          files, each containing all of the time steps for the entire run. If
          the model is run in serial model, only one file will be produced. If
          the model is run in MPI mode, then multiple files will be produced
          with the grid of the simulation sectioned off into one piece per
          file. No file conversion is attempted, but some pre-processing is
          performed on the coordinate grids.
        - The files are typically larger than 16 GB, so interpolation method 3
          is chosen for the time-varying variables. Interpolation method 0 is
          chosen for the constants (e.g. dV). The coordinate grid is not
          uniform in any manner, so a custom interpolator is required for both
          interpolation methods.
        - Adding this interpolator to the reader is a work in process via a
          collaboration with the GAMERA modeling team.
        '''
        def __init__(self, file_dir, variables_requested=[],
                     printfiles=False, filetime=False, gridded_int=True,
                     verbose=False, **kwargs):
            super(MODEL, self).__init__(**kwargs)
            self.modelname = 'GAMERA_GM'
            t0 = perf_counter()

            # first, check for file list, create if DNE
            list_file = file_dir + self.modelname + '_list.txt'
            time_file = file_dir + self.modelname + '_times.txt'
            self.times, self.pattern_files = {}, {}
            if not isfile(list_file) or not isfile(time_file):
                # collect filenames
                files = sorted(glob(file_dir+'*.h5'))
                data_files = [f for f in files if 'Res' not in f]
                self.filename = ''.join([f+',' for f in data_files])[:-1]
                # one pattern per run: abcd_00nx_00ny_00nz
                # abcd part might not always be four characters
                p = ''.join([b+'_' for b in basename(
                    data_files[0]).split('.')[0].split('_')[:-3]])[:-1]

                # establish time attributes
                # get list of files to loop through later
                self.pattern_files[p] = data_files
                self.times[p] = {'start': [], 'end': [], 'all': []}

                # all times are in each file, so just use first file
                h5_data = h5py.File(data_files[0])
                timestep_keys = [key for key in h5_data.keys()
                                 if 'Step' in key]
                mjd_list = [h5_data[key].attrs['MJD'] for key in
                            timestep_keys]
                h5_data.close()

                # convert from modified julian date to UTC time
                # output format is 'YYYY-MM-DD HH:MM:SS.mms
                t = sorted(Time(mjd_list, format='mjd').iso)
                self.filedate = datetime.fromisoformat(t[0][:10]).replace(
                    tzinfo=timezone.utc)  # date only
                # convert to hrs since midnight
                utc_ts = timestr_timestamp(t)
                hrs = (utc_ts - self.filedate.timestamp()) / 3600.
                # store in self.times dictionary
                self.times[p]['start'] = repeat([hrs[0]], len(data_files))
                self.times[p]['end'] = repeat([hrs[-1]], len(data_files))
                self.times[p]['all'] = array(hrs)

                # need to figure out if millisecond accuracy is needed
                test = (len(utc_ts) - 1)/(utc_ts[-1] - utc_ts[0])  # #steps/s
                if test >= 1.5:  # need ms timing if >= 1.5 timesteps/second
                    ms_timing = True
                else:
                    ms_timing = False

                # Pre-calculate the ranges of the coordinate grids.
                # Needed to generate a default uniform grid for plotting.
                print('Computing grid ranges...', end="")
                G.GridRanges(data_files)
                print('done.')

                # Pre-calculate the centers of the coordinate grids for each
                # file separately. Needed for interpolation later.
                print('Computing cell centers...', end="")
                G.ComputeCellCentersFile(data_files)
                print('done.')

                # create time list file if DNE
                RU.create_timelist(list_file, time_file, self.modelname,
                                   self.times, self.pattern_files,
                                   self.filedate, ms_timing=ms_timing)
            else:  # read in data and time grids from file list
                self.times, self.pattern_files, self.filedate, self.filename =\
                    RU.read_timelist(time_file, list_file)
            if filetime:
                return  # return times only

            # perform initial check on variables_requested list
            if len(variables_requested) > 0 and variables_requested != 'all':
                test_list = [value[0] for key, value in model_varnames.items()]
                err_list = [item for item in variables_requested if item
                            not in test_list]
                if len(err_list) > 0:
                    print('Variable name(s) not recognized:', err_list)
                for item in err_list:
                    variables_requested.remove(item)
                if len(variables_requested) == 0:
                    return

            # collect variable list (in attributes of datasets)
            p = list(self.pattern_files.keys())[0]  # only one pattern
            h5_data = h5py.File(self.pattern_files[p][0])
            key_list = [key for key in h5_data.keys() if 'Step' not in key and
                        key not in ['X', 'Y', 'Z']]  # skip coordinates
            step_list = list(h5_data['Step#0'].keys())
            var_list = key_list + step_list
            h5_data.close()

            if len(variables_requested) > 0 and variables_requested != 'all':
                gvar_list = [key for key, value in model_varnames.items()
                             if value[0] in variables_requested and
                             key in var_list]  # file variable names

                # check for variables requested but not available
                if len(gvar_list) != len(variables_requested):
                    err_list = [value[0] for key, value in
                                model_varnames.items() if value[0] in
                                variables_requested and key not in gvar_list]
                    if len(err_list) > 0:
                        print('Some requested variables are not available:',
                              err_list)
            else:  # only input variables on the avoid_list if requested
                gvar_list = [key for key in var_list if key in
                             model_varnames.keys()]
                # return list of variables included in data files
                if variables_requested == 'all':
                    self.var_dict = {value[0]: value[1:] for key, value in
                                     model_varnames.items()
                                     if key in gvar_list}
                    return

            # initialize data mapping for each variable desired
            self.variables = {model_varnames[var][0]:
                              {'units': model_varnames[var][-1],
                               'data': p} for var in gvar_list}

            # read in coordinate grid ranges and make sample grids
            # X, Y, and Z are the same shape in the data because they are the
            # x, y, and z coordinates of the same set of points.
            X_min, X_max, Y_min, Y_max, Z_min, Z_max =\
                G.Read_GridRanges(file_dir)
            if len(self.pattern_files[p]) > 1:
                sample_var = [gvar for gvar in model_varnames.keys()
                              if gvar in step_list][0]
                net_indices = G.GridMapping(self.pattern_files[p], False,
                                            sample_var)
                shape = net_indices['net'][1]  # shape of total array
            else:
                h5_data = h5py.File(self.pattern_files[p][0])
                shape = list(h5_data['X'].shape)  # shape of total array
                h5_data.close()
            self._X = linspace(X_min, X_max, endpoint=True, num=shape[0])
            self._Y = linspace(Y_min, Y_max, endpoint=True, num=shape[1])
            self._Z = linspace(Z_min, Z_max, endpoint=True, num=shape[2])

            # store a few items
            self.missing_value = NaN
            if verbose:
                print(f'Took {perf_counter()-t0:.6f}s to read in data')
            if printfiles:
                print(self.filename)

            # GAMERA data has all the timestamps in all the files.
            # This will confuse the lazy interpolation, so stripping down the
            # times dict to one file entry (only start and end fields needed).
            # For ms_timing, 'all' has ms resolution, start and end do not.
            self.singletimes = {'start': [self.times[p]['all'][0]],
                                'end': [self.times[p]['all'][-1]]}

            # register interpolators for each requested variable
            t_reg = perf_counter()
            # store original list b/c gridded interpolators change key list
            varname_list = [key for key in self.variables.keys()]
            for varname in varname_list:
                self.register_variables(varname, gridded_int)
            if verbose:
                print(f'Took {perf_counter()-t_reg:.5f}s to register ' +
                      f'{len(varname_list)} variables.')
            if verbose:
                print(f'Took a total of {perf_counter()-t0:.5f}s to kamodofy' +
                      f' {len(varname_list)} variables.')

        # define and register the variable
        def register_variables(self, varname, gridded_int):
            '''Functionalizes the indicated dataset.'''

            # determine coordinate variables and xvec by coord list
            key = self.variables[varname]['data']
            coord_list = [value[5] for key, value in model_varnames.items()
                          if value[0] == varname][0]
            gvar = [key for key, value in model_varnames.items() if
                    value[0] == varname][0]
            coord_str = [value[3]+value[4] for key, value in
                         model_varnames.items() if value[0] == varname][0]
            file_dir = self.pattern_files[key][0].split(
                basename(self.pattern_files[key][0]))[0]
            if 'time' in coord_list:
                coord_dict = {'time': {'units': 'hr',
                                       'data': self.times[key]['all']},
                              'X': {'units': 'R_E', 'data': self._X},
                              'Y': {'units': 'R_E', 'data': self._Y},
                              'Z': {'units': 'R_E', 'data': self._Z}}

                def func(i, fi):  # i = file# (always 0), fi = slice# (= Step#)
                    # Leave data in blocks. Need interpolator to determine
                    # which block is needed. Setting default of block 0 for now
                    block = 0  # default, but need this value from the interp
                    # read in data from block
                    file = self.pattern_files[key][block]
                    h5_data = h5py.File(file)
                    data = array(h5_data['Step#'+str(fi)][gvar])
                    h5_data.close()
                    # read in cell centers for requested block
                    center_file = file_dir + basename(file).split('.')[0] +\
                        '_gridcenters.nc'
                    cdf_data = Dataset(center_file)
                    X_c = array(cdf_data['X_c'])
                    Y_c = array(cdf_data['Y_c'])
                    Z_c = array(cdf_data['Z_c'])
                    cdf_data.close()

                    # define custom interpolator here  ***********************************
                    def interp(xvec):
                        tic = perf_counter()
                        X, Y, Z = xvec.T  # xvec can be used like this
                        # call custom interpolator here
                        print('variable', i, fi, data.shape, X_c.shape)
                        # dummy code for now
                        if not isinstance(X, ndarray):
                            return NaN
                        else:
                            return zeros(len(X)) * NaN
                    return interp

                # functionalize the variable dataset
                tmp = self.variables[varname]
                tmp['data'] = zeros((2, 2, 2, 2))  # saves execution time
                self = RU.Functionalize_Dataset(
                    self, coord_dict, varname, tmp, gridded_int, coord_str,
                    interp_flag=3, func=func, times_dict=self.singletimes,
                    func_default='custom')
                # choosing large chunked format, same as WACCMX.
            else:  # functionalize the constants
                coord_dict = {'X': {'units': 'R_E', 'data': self._X},
                              'Y': {'units': 'R_E', 'data': self._Y},
                              'Z': {'units': 'R_E', 'data': self._Z}}

                def func():
                    # Leave data in blocks. Need interpolator to determine
                    # which block is needed. Setting default of block 0 for now
                    block = 0  # default, but need this value from the interp
                    # read in data from block
                    file = self.pattern_files[key][block]
                    h5_data = h5py.File(file)
                    data = array(h5_data[gvar])
                    h5_data.close()
                    # read in cell centers for requested block
                    center_file = file_dir + basename(file).split('.')[0] +\
                        '_gridcenters.nc'
                    cdf_data = Dataset(center_file)
                    X_c = array(cdf_data['X_c'])
                    Y_c = array(cdf_data['Y_c'])
                    Z_c = array(cdf_data['Z_c'])
                    cdf_data.close()

                    # define custom interpolator here  ***********************************
                    def interp(xvec):
                        tic = perf_counter()
                        X, Y, Z = array(xvec).T  # xvec can be used like this
                        # call custom interpolator here
                        print('constant', data.shape, X_c.shape)
                        # dummy code for now
                        if not isinstance(X, ndarray):
                            return NaN
                        else:
                            return zeros(len(X)) * NaN
                    return interp

                # functionalize the variable dataset
                tmp = self.variables[varname]
                tmp['data'] = zeros((2, 2, 2))  # saves execution time

                self = RU.Functionalize_Dataset(
                    self, coord_dict, varname, tmp, gridded_int, coord_str,
                    interp_flag=0, func=func, func_default='custom')
            return
    return MODEL



# -*- coding: utf-8 -*-
"""
Created on Tue Mar 21 13:25:49 2023

@author: rringuet
"""
import numpy as np
from datetime import datetime
from kamodo import Kamodo
from kamodo_ccmc.tools.functionalize import Functionalize_Dataset


def varlist(meta):
    '''Creates a list of variables names as found in the meta object returned
    by HAPI.
    Example usage: var_list = varlist(meta)'''
    var_list = [meta['parameters'][i+1]['name'] for i in range(len(
        meta['parameters'])-1)]
    return var_list


@np.vectorize
def fromiso_totimestamp(iso):
    '''Converts an iso string into an UTC timestamp.
    Example usage: timestamp_arr = fromiso_totimestamp(iso_arr)'''
    return datetime.fromisoformat(iso).timestamp()


def time_data(data):
    '''Retrieves the ISO time strings from HAPI's meta object; returns UTC
    timestamps for Kamodo.
    Example usage: time_grid = time_data(data)'''
    return fromiso_totimestamp([str(val[0])[2:].split('Z')[0].strip("'") for
                                val in data])


def kamodo_units(unit_str):
    '''Tests that unit is supported in Kamodo. Returns the corrected unit
    string if yes, or an empty string if no.
    Example usage: units = test_units(unit_str)'''
    x = np.linspace(0, 1, 10)  # sample data
    u = unit_str.replace(' ', '*')  # can't replace in place must copy
    try:
        ko = Kamodo('T['+u+']=x')
        return u
    except:
        return ''


def functionalize_variable(coord_dict, var_dict, meta_par, kamodo_object=None,
                      custom_interp=None, coord_str=''):
    '''Functionalize the given variable with an interpolator.
    Inputs:
        coord_dict: a dictionary containing the coordinate information.
            {'name_of_coord1': {'units': 'coord1_units', 'data': coord1_data},
             'name_of_coord2': {'units': 'coord2_units', 'data': coord2_data},
             etc...}
            coordX_data should be a 1D array. All others should be strings.
        var_dict: a dictionary containing the data information.
            {'variable_name1': {'units': 'data1_units', 'data': data1_array},
             'variable_name2': {'units': 'data2_units', 'data': data2_array},
             etc...}
            dataX_array should have the same shape as
                (coord1, coord2, coord3, ..., coordN)
        meta_par: meta['parameters'][i+1] where meta is the meta object
            returned from HAPI
        kamodo_object: a object created by Kamodo
        custom_interp: a custom interpolator with the same execution syntax as
            SciPy's interp1d and RegularGridInterpolator functions. The command
            ```py
            interp = custom_interp(coord_dict, var_dict)
            ```
            must initialize the interpolator with the given coordinate
            dictionary and the variable dictionary as described above.
            Default is none to functionalize with the standard interpolator.
        coord_str: a string indicating the coordinate system
            (e.g. 'GEOsph' for spherical geographical coordinates).
            See Kamodo's ConvertCoord function for more information.
            Default is an empty string.
    Returns a kamodo object with the variable added to it.
    Called by kamodofy_hapi function below.
    '''
    var_name = list(var_dict.keys())[0]
    if len(var_dict[var_name]['data'].shape) == 1:  # coord_dict fine as is
        if custom_interp is None:
            kamodo_object = Functionalize_Dataset(
                coord_dict, var_dict, kamodo_object, coord_str=coord_str)
        else:
            # initialize custom interpolation method with the variable data
            interp = custom_interp(coord_dict, var_dict)
            # functionalize
            kamodo_object = Functionalize_Dataset(
                coord_dict, var_dict, kamodo_object, coord_str=coord_str,
                func=interp, func_default='custom')
    elif len(var_dict[var_name]['data'].shape) > 1:  # add more coordinates
        coord_names = [meta_par['bins'][j]['name'] for j in
                       range(len(meta_par['bins']))]
        for j, c in enumerate(coord_names):
            coord_u = kamodo_units(meta_par['bins'][j]['units'])
            coord_dict[c] = {'data': np.array(meta_par['bins'][j]['centers']),
                             'units': coord_u}
            if len(coord_dict[c]['data'].shape) > 1:
                print('Coordinate name {c}, coordinate dimensions ' +
                      f'{coord_dict[c]["data"].shape}')
                print('Coordinate arrays of dimension higher than 1D require' +
                      ' a custom interpolator from the user. ' +
                      'See documentation for how this can be done.')
                return kamodo_object
        if custom_interp is None:
            kamodo_object = Functionalize_Dataset(
                coord_dict, var_dict, kamodo_object, coord_str=coord_str)
        else:
            # initialize custom interpolation method with the variable data
            interp = custom_interp(coord_dict, var_dict)
            # functionalize
            kamodo_object = Functionalize_Dataset(
                coord_dict, var_dict, kamodo_object, coord_str=coord_str,
                func=interp, func_default='custom')
    return kamodo_object


def functionalize_hapi(data, meta, custom_interp=None, coord_str=''):
    '''Functionalizes the data found in the data object given.
    Inputs:
        data: data object returned by HAPI
        meta: meta object returned by HAPI
        custom_interp: a custom interpolator with the same execution syntax as
            SciPy's interp1d and RegularGridInterpolator functions. If a list
            is given, the interpolators should be given in the same order as
            in the meta object. Default is None.
        custom_interp: a custom interpolator with the same execution syntax as
            SciPy's interp1d and RegularGridInterpolator functions. If a list
            is given, the interpolators should be given in the same order as
            in the meta object. The command
            ```py
            interp = custom_interp(coord_dict, var_dict)
            ```
            must initialize the interpolator with the given coordinate
            dictionary and the variable dictionary.
            Default is none to functionalize with the standard interpolator.
            Required inputs:
                coord_dict: a dictionary containing the coordinate information.
                    {'c1_name': {'units': 'c1_units', 'data': c1_data},
                     'c2_name': {'units': 'c2_units', 'data': c2_data},
                     etc...}
                    cX_data should be a 1D array. cX_name should be the desired
                    LaTeX representation of the coordinate name. cX_units
                    should be a string representing the unit to be
                    functionalized in Kamodo.
                var_dict: a dictionary containing the data information.
                    {'var_name1': {'units': 'var1_units', 'data': var1_data},
                     'var_name2': {'units': 'var2_units', 'data': var2_data},
                     etc...}
                    varX_data arrays should have the same shape as the
                    coordinates given in coord_dict
                    (e.g. (c1, c2, c3, ..., cN)).
                These objects are typically created with the following code:
                    ```py
                    time_grid = time_data(data)
                    var_list = varlist(meta)
                    for i, key in enumerate(var_list):
                        u = test_units(meta['parameters'][i+1]['units'])
                        var_dict = {key: {'data': np.array([val[i+1] for val in
                                                            data]),
                                          'units': u}}
                        coord_dict = {'UTC_time': {'units': 's',
                                                   'data': time_grid},
                                      'c1': {'units': 'R_E', 'data': .....},
                                      etc...}
                    ```
                where time_data and varlist are functions in the
                functionalize_hapi script. See the functionalize_hapi function
                logic in the same script for more details.
        coord_str: a string indicating the coordinate system
            (e.g. 'GEOsph' for spherical geographical coordinates).
            See Kamodo's ConvertCoord function for more information.
            Default is an empty string.
    Returns:
        A kamodo object with all possible variables functionalized.
    '''

    time_grid = time_data(data)  # convert iso strings to UTC timestamps
    var_list = varlist(meta)  # also used later
    kamodo_object = None  # initialize for the loop
    for i, key in enumerate(var_list):
        u = kamodo_units(meta['parameters'][i+1]['units'])  # check units Kamodo
        var_dict = {key: {'data': np.array([val[i+1] for val in data]),
                          'units': u}}
        coord_dict = {'UTC_time': {'units': 's', 'data': time_grid}}
        if isinstance(custom_interp, list):
            interp = custom_interp[i]
        else:
            interp = custom_interp
        kamodo_object = functionalize_variable(
            coord_dict, var_dict, meta['parameters'][i+1], kamodo_object,
            custom_interp=interp, coord_str=coord_str)
    return kamodo_object



'''
Written by Rebecca Ringuette, 2021
'''
from datetime import datetime, timedelta, timezone

# variable name in file: [standardized variable name, descriptive term, units]
model_varnames = {"V": ['V', 'Electric potential', 0, 'SM', 'sph',
                        ['time', 'lon', 'lat'], 'kV'],
                  "theta_v": ['theta_v', 'Azimuthal angle of convection ' +
                              'velocity', 1, 'SM', 'sph', ['time', 'lon',
                                                           'lat'], 'deg'],
                  "v": ['v', 'magnitude of convection velocity', 2, 'SM',
                        'sph', ['time', 'lon', 'lat'], 'm/s'],
                  # remaining variables are time series
                  "theta_Btilt": ['theta_Btilt', 'Dipole tilt', 3, 'SM',
                                  'sph', ['time'], 'deg'],
                  'E_sw': ['E_sw', 'Electric field of the solar wind', 4,
                           'SM', 'sph', ['time'], 'mV/m'],
                  'theta_B': ['theta_B', 'IMF clock angle', 5, 'SM', 'sph',
                              ['time'], 'deg']}


# times from file converted to seconds since midnight of filedate
# plotting input times will be datetime strings of format 'YYYY-MM-DD HH:mm:ss'
# filedate is self.filedate from superdarn object
# converts to hours since midnight of filedate for plotting
def MODEL():

    from kamodo import Kamodo
    from netCDF4 import Dataset
    from glob import glob
    from os.path import basename, isfile
    from numpy import array, NaN, unique
    from time import perf_counter
    import kamodo_ccmc.readers.reader_utilities as RU

    class MODEL(Kamodo):
        '''SuperDARN model data reader for the uniform coordinate grid.

        Inputs:
            file_dir: a string representing the file directory of the
                model output data.
                Note: This reader 'walks' the entire dataset in the directory.
            variables_requested = a list of variable name strings chosen from
                the model_varnames dictionary in this script, specifically the
                first item in the list associated with a given key.
                - If empty, the reader functionalizes all possible variables
                    (default)
                - If 'all', the reader returns the model_varnames dictionary
                    above for only the variables present in the given files.
            filetime = boolean (default = False)
                - If False, the script fully executes.
                - If True, the script only executes far enough to determine the
                    time values associated with the chosen data.
            printfiles = boolean (default = False)
                - If False, the filenames associated with the data retrieved
                    ARE NOT printed.
                - If True, the filenames associated with the data retrieved ARE
                    printed.
            gridded_int = boolean (default = True)
                - If True, the variables chosen are functionalized in both the
                    standard method and a gridded method.
                - If False, the variables chosen are functionalized in only the
                    standard method.
            verbose = boolean (False)
                - If False, script execution and the underlying Kamodo
                    execution is quiet except for specified messages.
                - If True, be prepared for a plethora of messages.
        All inputs are described in further detail in
            KamodoOnboardingInstructions.pdf.

        Returns: a kamodo object (see Kamodo core documentation) containing all
            requested variables in functionalized form.

        Notes:
            - The SuperDARN default grid model output is given in one ascii
              file per timestep per N/S hemisphere. The file converter combines
              these files into one netCDF4 file per timestep.
            - The data is only given within 40-60 degrees of the poles, so a
              buffer row of the same values is added in the data to avoid
              losing the most equatorward ring of data in the interpolation.
            - The outputs do not provide values at the poles, so scalar
              averaging is used to determine these values. (The three
              spatially dependent variables are all scalars.)
            - The converted files are small and are created with one time step
              per file, so interpolation method 1 is chosen for spatially
              dependent variables. Interpolation method 0 is chosen for the
              spatially independent variables, meaning the entire time series
              is read into memory and the standard SciPy interpolator is used.
              For the spatially dependent variables, the standard SciPy
              interpolator is used, but with a custom layer to accomodate the
              changing range of the latitude grid with time.
        '''
        def __init__(self, file_dir, variables_requested=[],
                     filetime=False, verbose=False, gridded_int=True,
                     printfiles=False, **kwargs):
            super(MODEL, self).__init__(**kwargs)
            self.modelname = 'SuperDARN_uni'
            t0 = perf_counter()

            # first, check for file list, create if DNE
            list_file = file_dir + self.modelname + '_list.txt'
            time_file = file_dir + self.modelname + '_times.txt'
            self.times, self.pattern_files = {}, {}
            if not isfile(list_file) or not isfile(time_file):
                # make sure all files are converted
                from kamodo_ccmc.readers.superdarn_tocdf import \
                    convert_all
                tmp = convert_all(file_dir, '*uni.txt')

                # continue with converted files
                files = sorted(glob(file_dir+'*_uni.nc'))
                self.filename = ''.join([f+',' for f in files])[:-1]
                patterns = unique([basename(f)[:-20] for f in files])
                self.filedate = datetime.strptime(
                    basename(files[0])[-20:-12], '%Y%m%d').replace(
                        tzinfo=timezone.utc)

                # establish time attributes from filenames
                for p in patterns:
                    # get list of files to loop through later
                    pattern_files = sorted(glob(file_dir+p+'*_uni.nc'))
                    self.pattern_files[p] = pattern_files
                    self.times[p] = {'start': [], 'end': [], 'all': []}
                    if pattern_files[0][-12] == '-':
                        format_string = '%Y%m%d-%H%M'
                    else:
                        format_string = '%Y%m%d_%H%M'

                    # loop through to get times, one file per time step
                    self.times[p]['start'] = array([
                        RU.str_to_hrs(f[-20:-7], self.filedate,
                                      format_string=format_string) for f in
                        pattern_files])
                    self.times[p]['end'] = self.times[p]['start'].copy()
                    self.times[p]['all'] = self.times[p]['start'].copy()

                # create time list file if DNE
                RU.create_timelist(list_file, time_file, self.modelname,
                                   self.times, self.pattern_files,
                                   self.filedate)
            else:  # read in data and time grids from file list
                self.times, self.pattern_files, self.filedate, self.filename =\
                    RU.read_timelist(time_file, list_file)
            if filetime:
                return  # return times
            # only one pattern, so simplifying code
            p = list(self.pattern_files.keys())[0]
            cdf_data = Dataset(self.pattern_files[p][0])

            # perform initial check on variables_requested list
            if len(variables_requested) > 0 and variables_requested != 'all':
                test_list = [value[0] for key, value in model_varnames.items()]
                err_list = [item for item in variables_requested if item not in
                            test_list]
                if len(err_list) > 0:
                    print('Variable name(s) not recognized:', err_list)
                for item in err_list:
                    variables_requested.remove(item)
                if len(variables_requested) == 0:
                    return

            # collect variable list
            if len(variables_requested) > 0 and variables_requested != 'all':
                gvar_list = [key for key, value in model_varnames.items()
                             if value[0] in variables_requested and
                             key in cdf_data.variables.keys()]

                # check for variables requested but not available
                if len(gvar_list) != len(variables_requested):
                    err_list = [value[0] for key, value in
                                model_varnames.items() if value[0] in
                                variables_requested and key not in gvar_list]
                    if len(err_list) > 0:
                        print('Some requested variables are not available:',
                              err_list)
            else:  # only input variables on the avoid_list if requested
                gvar_list = [key for key in cdf_data.variables.keys()
                             if key in model_varnames.keys()]
                # returns list of variables included in data files
                if variables_requested == 'all':
                    self.var_dict = {value[0]: value[1:] for key, value in
                                     model_varnames.items() if key in
                                     gvar_list}
                    cdf_data.close()
                    return

            # store mapping for each variable desired
            self.variables = {model_varnames[var][0]: {
                'units': model_varnames[var][-1], 'data': p} for var in
                gvar_list}

            # Store coordinate data as class attributes
            self._lon = array(cdf_data.variables['lon'])  # -180 to 180
            self._lat = array(cdf_data.variables['lat'])
            cdf_data.close()

            # store a few items
            self.missing_value = NaN
            self._registered = 0
            if verbose:
                print(f'Took {perf_counter()-t0:.6f}s to read in data')
            if printfiles:
                print(self.filename)

            # register interpolators for each requested variable
            t_reg = perf_counter()
            # store original list b/c gridded interpolators change key list
            varname_list = [key for key in self.variables.keys()]
            for varname in varname_list:
                self.register_variable(varname, gridded_int)

            if verbose:
                print(f'Took {perf_counter()-t_reg:.5f}s to register ' +
                      f'{len(varname_list)} variables.')
            if verbose:
                print(f'Took a total of {perf_counter()-t0:.5f}s to kamodofy' +
                      f' {len(varname_list)} variables.')
            return

        # define and register a variable
        def register_variable(self, varname, gridded_int):
            """Registers an interpolator with proper signature"""

            # determine coordinate variables and xvec by coord list
            key = self.variables[varname]['data']
            coord_list = [value[5] for key, value in model_varnames.items()
                          if value[0] == varname][0]
            coord_dict = {'time': {'units': 'hr',
                                   'data': self.times[key]['all']}}
            gvar = [key for key, value in model_varnames.items() if
                    value[0] == varname][0]
            coord_str = [value[3]+value[4] for key, value in
                         model_varnames.items() if value[0] == varname][0]

            # retrieve coordinate grids
            if len(coord_list) == 1:  # read in entire time series
                data = []
                for f in self.pattern_files[key]:
                    cdf_data = Dataset(f)
                    tmp = array(cdf_data.variables[gvar])[0]
                    cdf_data.close()
                    data.append(tmp)  # one value per file
                self.variables[varname]['data'] = array(data)
                self = RU.Functionalize_Dataset(
                    self, coord_dict, varname, self.variables[varname],
                    gridded_int, coord_str, interp_flag=0)  # single array
                return

            if len(coord_list) == 3:
                coord_dict['lon'] = {'units': 'deg', 'data': self._lon}
                coord_dict['lat'] = {'units': 'deg', 'data': self._lat}

                # define operations for each variable when given the key
                def func(i):
                    '''i is the file/time number.'''
                    # get data from file
                    file = self.pattern_files[key][i]
                    cdf_data = Dataset(file)
                    data = array(cdf_data.variables[gvar])
                    lat = array(cdf_data.variables['lat'])
                    cdf_data.close()
                    # data wrangling all done in the file conversion step
                    coord_list = [coord_dict['lon']['data'], lat]
                    interp = RU.rgiND(coord_list, data, bounds_error=False,
                                      fill_value=NaN)
                    return interp

                # functionalize the variable dataset
                self = RU.Functionalize_Dataset(
                    self, coord_dict, varname, self.variables[varname],
                    gridded_int, coord_str, interp_flag=1, func=func,
                    func_default='custom')
                return
    return MODEL


def coupling():
    '''Write output to go into the SuperDARN model.'''
    return



# Collaboration Example
# ---------------------
# This script outlines the items needed from the modeling team to collaborate
# on getting the model outputs accessible through Kamodo. This includes the
# model_varnames dictionary, the custom interpolation routine, the custom
# coordinate conversion routine (if needed), and a coordinate grid in both the
# standard and custom coordinate systems for plotting purposes.


# Below is an example of the model_varnames metadata dictionary. The keys are
# the names of the variables as given in the files. The list begins with the
# LaTeX representation of the variable and is followed by a description of the
# variable. These are followed by an integer, coordinate information, a list of
# the coordinate names, and the units. See the ConcertCoord function in
# kamodo_ccmc/flythrough/utils.py for more details on the coordinate systems.
# See https://ensemblegovservices.github.io/kamodo-core/notebooks/Syntax/
# for details on naming syntax.
# In the example below, '_ilev1' is added to LaTeX representation of the
# variable that depend on pressure level (the model specific coordinate system)
# to distinguish from the version of the variable that depends on the standard
# coordinate system.
model_varnames = {'rho': ['rho_ilev1', 'total mass density', 0, 'GDZ', 'sph',
                          ['time', 'lon', 'lat', 'ilev1'], 'kg/m**3'],
                  'rho_2': ['rho', 'total mass density', 0, 'GDZ', 'sph',
                            ['time', 'lon', 'lat', 'height'], 'kg/m**3'],
                  'T': ['T_ilev1', 'temperature', 1, 'GDZ', 'sph',
                        ['time', 'lon', 'lat', 'ilev1'], 'K'],
                  'T_2': ['T', 'temperature', 1, 'GDZ', 'sph',
                          ['time', 'lon', 'lat', 'height'], 'K'],
                  'T_e': ['T_e', 'electron temperature', 2, 'GDZ', 'sph',
                          ['time', 'lon', 'lat', 'height'], 'K'],
                  'T_i': ['T_i', 'ion temperature', 3, 'GDZ', 'sph',
                          ['time', 'lon', 'lat', 'height'], 'K'],
                  'H_ilev': ['H_ilev',
                             'height dependent on primary pressure level',
                             4, 'GDZ', 'sph', ['time', 'lon', 'lat', 'ilev'],
                             'm'],
                  'H_lev': ['H_ilev1',
                            'height dependent on secondary pressure level',
                            5, 'GDZ', 'sph', ['time', 'lon', 'lat', 'ilev1'],
                            'm']}

# If the file format is binary, compressed, or otherwise difficult to access,
# we need a function that returns a list of the variable names that are
# available in the chosen file(s).
# The beginning timestamp of each file should be calculable from the filename.
# If more than one timestamp is present in each file, and the file format is
# difficult to access, then we need a funtion that returns the first and last
# timestamp for the file or the interpolating range set in the custom
# interpolator, and the time resolution of the data.

# For each type of coordinate, we will need the associated unit given in a
# separate dictionary as in the example below.
coord_dict = {'time': 'hr', 'lon': 'deg', 'lat': 'deg', 'ilev': 'm/m',
              'height': 'km', 'ilev1': 'm/m', 'X': 'R_E', 'Y': 'R_E',
              'Z': 'R_E'}  # several examples are included here


# The 'interp_library' function below (line 72) should return the interpolating
# function for the given variable. The returned interpolating function should
# accept inputs of the form of the standard_track variable in the example
# above, either in numpy array or list form. The return value of the
# interpolation function should be the interpolated value in the MODEL-SPECIFIC
# coordinate system. E.g. value = interp(custom_track)

# Example track
import numpy as np
t = np.linspace(0., 24., 25)
c1 = np.linspace(-5, 25, 15)
c2 = np.linspace(-25, 2, 20)
c3 = np.linspace(-15, 15, 11)
tt, cc1, cc2, cc3 = np.meshgrid(t, c1, c2, c3)
t, c1, c2, c3 = np.ravel(tt), np.ravel(cc1), np.ravel(cc2), np.ravel(cc3)
standard_track = np.array([t, c1, c2, c3]).T
standard_track.shape  # (82500, 4)
# each row of the array/list should be a coordinate set
# e.g. [t_val, c1_val, c2_val, c3_val] for each row


# The names of the functions and script can be chosen by the developing team
# and should reflect the purpose of the  function/script. The name of the
# script should also begin with the name of the model.
def CustomInterp(variable_name):
    '''Custom Interpolator for given variable name in the model-specific
    coordinate system.'''
    # import custom interpolator script, likely via f2py interface
    from model_special import interp_library
    # call interpolator function 
    interp = interp_library(variable_name)
    return interp

# Include an example of how to use the interpolating function
T_interp = CustomInterp('T')
T_values = T_interp(custom_track)  # should return a 1D array or list of values
# custom_track should be an array similar to standard_track but with
# model-specific coordinate values of shape
# (number of positions, number of coordinate values per position)
# The interpolator should cover the entirety of the model grid with the
# exception of time, which can be sectioned off into single or multi-hour
# segments (preferably 24-hour segments if it will fit into 15 GB of memory).
# If the time grid is sectioned off, then the interpolator should
# properly handle interpolations between the time sections.


def CustomCoordConv():
    '''Custom function converting from a satellite track in a standard
    coordinate system to a track in the model-specific coordinate system.'''
    # import custom coordinate conversion script, likely via f2py interface
    from model_special import coord_conv
    return coord_conv

# Include an example of how to use the function, both with and without a 
# variable-specific interpolator
coord_conv = CustomCoordConv()
custom_track = coord_conv(standard_track)
# should return the values of the model-specific
# coordinates for the given coordinates in a standard system.
# def test(t, standard_coord1, standard_coord2, standard_coord3):
#   ---convert coordinates here----
#   return np.array([t, custom_coord1, custom_coord2, custom_coord3, ...])

# The two functions should be able to be used together, as below.
T_values = T_interp(coord_conv(standard_track))


# Finally, I need a set of 1D arrays with default values for the model-specific
# coordinate system. The max and min of the arrays should correspond to the max
# and min of the grid in standard coordinates. These will be used as the
# default coordinate grids for plotting. If these are different for each run,
# then a function should be created to return them, such as the one below.
def Coord_Ranges():
    # import custom script, likely via f2py
    from model_special import coord_ranges
    t, c1, c2, c3 = coord_ranges()
    return t, c1, c2, c3  # can add additional coordinate dimensions as needed


# We will work together on getting the gridded versions of these functions
# working once I get the reader together. Plotting in kamodo is problematic
# at best until the gridded versions of these functions are working.
